# This code was copied from [soynlp](https://github.com/lovit/soynlp)
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import re
from functools import partial
from typing import List, Tuple, Union

from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_num_workers, _check_text, _check_type

kor_begin = 44032
kor_end = 55203
chosung_base = 588
jungsung_base = 28
jaum_begin = 12593
jaum_end = 12622
moum_begin = 12623
moum_end = 12643

chosung_list = ['ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ',
                'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']
jungsung_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ',
                 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ',
                 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ',
                 'ㅡ', 'ㅢ', 'ㅣ']
jongsung_list = [
    ' ', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ',
    'ㄹ', 'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ',
    'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ',
    'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']
jaum_list = ['ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ', 'ㄸ', 'ㄹ',
             'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 'ㅁ', 'ㅂ',
             'ㅃ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']
moum_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ',
             'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ']

repeatchars_pattern = re.compile('(\w)\\1{2,}')


def _compose(chosung, jungsung, jongsung):
    return chr(kor_begin + chosung_base * chosung_list.index(chosung) + jungsung_base * jungsung_list.index(
        jungsung) + jongsung_list.index(jongsung))


def _decompose(c):
    if not _is_hangul(c):
        return None
    i = _to_base(c)
    if jaum_begin <= i <= jaum_end:
        return c, ' ', ' '
    if moum_begin <= i <= moum_end:
        return ' ', c, ' '
    i -= kor_begin
    cho = i // chosung_base
    jung = (i - cho * chosung_base) // jungsung_base
    jong = (i - cho * chosung_base - jung * jungsung_base)
    return chosung_list[cho], jungsung_list[jung], jongsung_list[jong]


def _is_hangul(c):
    i = _to_base(c)
    return (kor_begin <= i <= kor_end) or (jaum_begin <= i <= jaum_end) or (moum_begin <= i <= moum_end)


def _is_composed_hangul(c):
    return kor_begin <= _to_base(c) <= kor_end


def _is_consonant(c):
    return jaum_begin <= _to_base(c) <= jaum_end


def _is_vowel(c):
    return moum_begin <= _to_base(c) <= moum_end


def _to_base(c):
    if type(c) == str or type(c) == int:
        return ord(c)
    else:
        raise TypeError


def _character_is_number(i):
    i = _to_base(i)
    return 48 <= i <= 57


def _character_is_english(i):
    i = _to_base(i)
    return (97 <= i <= 122) or (65 <= i <= 90)


def _character_is_punctuation(i):
    i = _to_base(i)
    return i == 33 or i == 34 or i == 39 or i == 44 or i == 46 or i == 63 or i == 96


def reduce_char_repeats(
    text: Union[str, List[str], Tuple[str]],
    num_repeats: int = 2,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This reduces character repeats in text.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        num_repeats (int): the number of character that can be repeated
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: text with reduced character repeats or list of texts with reduced character repeats

    Examples:
        >>> from kss import Kss
        >>> reduce_char_repeats = Kss("reduce_char_repeats")
        >>> text = "고고고고고고고"
        >>> output = reduce_char_repeats(text)
        >>> print(output)
        '고고'

    References:
        This was copied from [soynlp](https://github.com/lovit/soynlp) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_repeats = _check_type(num_repeats, "num_repeats", int)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(_reduce_char_repeats, num_repeats=num_repeats),
        inputs=text,
        num_workers=num_workers,
    )


def reduce_emoticon_repeats(
    text: Union[str, List[str], Tuple[str]],
    num_repeats: int = 2,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This reduces emoticon repeats in text.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        num_repeats (int): the number of emoticon that can be repeated
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: text with reduced emoticon repeats or list of texts with reduced emoticon repeats

    Examples:
        >>> from kss import Kss
        >>> reduce_emoticon_repeats = Kss("reduce_emoticon_repeats")
        >>> text = "앜ㅋㅋㅋㅋㅋㅋ"
        >>> output = reduce_emoticon_repeats(text)
        >>> print(output)
        '아ㅋㅋ'

    References:
        This was copied from [soynlp](https://github.com/lovit/soynlp) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_repeats = _check_type(num_repeats, "num_repeats", int)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(_reduce_emoticon_repeats, num_repeats=num_repeats),
        inputs=text,
        num_workers=num_workers,
    )


def _reduce_char_repeats(sent, num_repeats=2):
    if num_repeats > 0:
        sent = repeatchars_pattern.sub('\\1' * num_repeats, sent)
    return sent


def _reduce_emoticon_repeats(sent, num_repeats=2):
    if not sent:
        return sent

    # Pattern matching ㅋ쿠ㅜ
    def pattern(idx):
        # Jaum: 0, Moum: 1, Complete: 2, else -1
        if 12593 <= idx <= 12622:
            return 0
        elif 12623 <= idx <= 12643:
            return 1
        elif 44032 <= idx <= 55203:
            return 2
        else:
            return -1

    idxs = [pattern(ord(c)) for c in sent]
    sent_ = []
    last_idx = len(idxs) - 1
    for i, (idx, c) in enumerate(zip(idxs, sent)):
        if (0 < i < last_idx) and (idxs[i - 1] == 0 and idx == 2 and idxs[i + 1] == 1):
            cho, jung, jong = _decompose(c)
            if (cho == sent[i - 1]) and (jung == sent[i + 1]) and (jong == ' '):
                sent_.append(cho)
                sent_.append(jung)
            else:
                sent_.append(c)
        elif (i < last_idx) and (idx == 2) and (idxs[i + 1] == 0):
            cho, jung, jong = _decompose(c)
            if jong == sent[i + 1]:
                sent_.append(_compose(cho, jung, ' '))
                sent_.append(jong)
        elif (i > 0) and (idx == 2 and idxs[i - 1] == 0):
            cho, jung, jong = _decompose(c)
            if cho == sent[i - 1]:
                sent_.append(cho)
                sent_.append(jung)
        else:
            sent_.append(c)
    return reduce_char_repeats(''.join(sent_), num_repeats)



# -*- coding: utf-8 -*-
from kss._modules.hangulization.hangulize import *


class Georgian(Language):
    """For transcribing Georgian (broad transcription).
    Only the lenis and aspirated series of stops and affricates of Korean are
    used. The Georgian grapheme ვ is consistently taken to be /v/, except after
    a vowel and before a voiceless stop or affricate, in which case it is taken
    to be /f/."""

    __iso639__ = {1: 'ka', 2: 'geo', 3: 'kat'}
    __tmp__ = ',;'

    vowels = u'აეიოუ'
    cs = u'ბგდვზთკლმნპჟრსტფქღყშჩცძწჭხჯჰჶ'
    vl = u'თკპტფქყჩცწჭ'
    ob = u'ბგდვზთკპჟსტფქღყშჩცძწჭხჯჰ'
    notation = Notation([
        (u'ჱ', u'ეჲ'),
        (u'ჲ', u'ი'),
        (u'უჳ', u'უ'),
        (u'ჳ', u'უ'),
        (u'ჴ', u'ხ'),
        (u'ჵ', u'ო'),
        (u'{@}ვ{<vl>}', u'ჶ'),
        (u'ვ$', u'ჶ'),
        (u'ბბ', u'ბ'),
        (u'გგ', u'გ'),
        (u'დდ', u'დ'),
        (u'ვვ', u'ვ'),
        (u'ზზ', u'ზ'),
        (u'თთ', u'თ'),
        (u'კკ', u'კ'),
        (u'ლლ', u'ლ'),
        (u'მმ', u'მ,მ'),
        (u'ნნ', u'ნ,ნ'),
        (u'პპ', u'პ'),
        (u'ჟჟ', u'ჟ'),
        (u'რრ', u'რ'),
        (u'სს', u'ს'),
        (u'ტტ', u'ტ'),
        (u'ფფ', u'ფ'),
        (u'ქქ', u'ქ'),
        (u'ღღ', u'ღ'),
        (u'ყყ', u'ყ'),
        (u'შშ', u'შ'),
        (u'ხხ', u'ხ'),
        (u'ჰჰ', u'ჰ'),
        (u'დ{ძ|ჯ}', None),
        (u'თ{ჩ|ც}', None),
        (u'ტ{წ|ჭ}', None),
        (u'დჟ', u'ჯ'),
        (u'თშ', u'ჩ'),
        (u'ტშ', u'ჭ'),
        (u'დზ', u'ძ'),
        (u'თს', u'ც'),
        (u'ტს', u'წ'),
        (u'ჟ{<cs>}', u'ჟუ'),
        (u'ჟ$', u'ჟუ'),
        (u'შ{<cs>}', u'შუ'),
        (u'შ$', u'ში'),
        (u'ჩ{ვ}', u'ჩუ'),
        (u'ჩ{<cs>}', u'ჩი'),
        (u'ჩ$', u'ჩი'),
        (u'ძ{ვ}', u'ძუ'),
        (u'ძ{<cs>}', u'ძი'),
        (u'ძ$', u'ძი'),
        (u'ჭ{ვ}', u'ჭუ'),
        (u'ჭ{<cs>}', u'ჭი'),
        (u'ჭ$', u'ძი'),
        (u'^ლ', u'ლ;'),
        (u'^ნ', u'ნ;'),
        (u'ლ$', u'ლ,'),
        (u'მ$', u'მ,'),
        (u'ნ$', u'ნ,'),
        (u'ლ{@|მ,|ნ,}', u'ლ;'),
        (u'{,}ლ', u'ლ;'),
        (u'მ{@}', u'მ;'),
        (u'ნ{@}', u'ნ;'),
        (u'ლ', u'ლ,'),
        (u'მ', u'მ,'),
        (u'ნ', u'ნ,'),
        (u',,', u','),
        (u',;', None),
        (u',ლ,', u'ლ,'),
        (u',მ,', u'მ,'),
        (u',ნ,', u'ნ,'),
        (u'ლ{მნ}', u'ლ,'),
        (u';', None),
        (u'აა', u'ა'),
        (u'ეე', u'ე'),
        (u'იი', u'ი'),
        (u'ოო', u'ო'),
        (u'უუ', u'უ'),
        (u'ბ', Choseong(B)),
        (u'გ', Choseong(G)),
        (u'დ', Choseong(D)),
        (u'ვ', Choseong(B)),
        (u'ზ', Choseong(J)),
        (u'თ', Choseong(T)),
        (u'კ', Choseong(K)),
        (u'^ლ', Choseong(L)),
        (u'{,}ლ', Choseong(L)),
        (u'ლ,', Jongseong(L)),
        (u'ლ', Jongseong(L), Choseong(L)),
        (u'მ,', Jongseong(M)),
        (u'მ', Choseong(M)),
        (u'ნ,', Jongseong(N)),
        (u'ნ', Choseong(N)),
        (u'პ', Choseong(P)),
        (u'ჟ', Choseong(J)),
        (u'რ', Choseong(L)),
        (u'ს', Choseong(S)),
        (u'ტ', Choseong(T)),
        (u'ფ', Choseong(P)),
        (u'ქ', Choseong(K)),
        (u'ღ', Choseong(G)),
        (u'ყ', Choseong(K)),
        (u'ჩ', Choseong(C)),
        (u'ც', Choseong(C)),
        (u'ძ', Choseong(J)),
        (u'წ', Choseong(C)),
        (u'ჭ', Choseong(C)),
        (u'ხ', Choseong(H)),
        (u'ჯ', Choseong(J)),
        (u'ჰ', Choseong(H)),
        (u'ჶ', Choseong(P)),
        (u'ჸ', Choseong(NG)),
        (u'შა', Choseong(S), Jungseong(YA)),
        (u'შე', Choseong(S), Jungseong(YE)),
        (u'ში', Choseong(S), Jungseong(I)),
        (u'შო', Choseong(S), Jungseong(YO)),
        (u'შუ', Choseong(S), Jungseong(YU)),
        (u'შჷ', Choseong(S), Jungseong(YEO)),
        (u'ა', Jungseong(A)),
        (u'ე', Jungseong(E)),
        (u'ი', Jungseong(I)),
        (u'ო', Jungseong(O)),
        (u'უ', Jungseong(U)),
        (u'ჷ', Jungseong(EO)),
    ])

    def normalize(self, string):
        return normalize_roman(string, {
            u'Ⴀ': u'ა', u'ⴀ': u'ა', u'Ⴁ': u'ბ', u'ⴁ': u'ბ', u'Ⴂ': u'გ',
            u'ⴂ': u'გ', u'Ⴃ': u'დ', u'ⴃ': u'დ', u'Ⴄ': u'ე', u'ⴄ': u'ე',
            u'Ⴅ': u'ვ', u'ⴅ': u'ვ', u'Ⴆ': u'ზ', u'ⴆ': u'ზ', u'Ⴡ': u'ჱ',
            u'ⴡ': u'ჱ', u'Ⴇ': u'თ', u'ⴇ': u'თ', u'Ⴈ': u'ი', u'ⴈ': u'ი',
            u'Ⴉ': u'კ', u'ⴉ': u'კ', u'Ⴊ': u'ლ', u'ⴊ': u'ლ', u'Ⴋ': u'მ',
            u'ⴋ': u'მ', u'Ⴌ': u'ნ', u'ⴌ': u'ნ', u'Ⴢ': u'ჲ', u'ⴢ': u'ჲ',
            u'Ⴍ': u'ო', u'ⴍ': u'ო', u'Ⴎ': u'პ', u'ⴎ': u'პ', u'Ⴏ': u'ჟ',
            u'ⴏ': u'ჟ', u'Ⴐ': u'რ', u'ⴐ': u'რ', u'Ⴑ': u'ს', u'ⴑ': u'ს',
            u'Ⴒ': u'ტ', u'ⴒ': u'ტ', u'Ⴣ': u'ჳ', u'ⴣ': u'ჳ', u'Ⴍ': u'უ',
            u'ⴍ': u'უ', u'Ⴔ': u'ფ', u'ⴔ': u'ფ', u'Ⴕ': u'ქ', u'ⴕ': u'ქ',
            u'Ⴖ': u'ღ', u'ⴖ': u'ღ', u'Ⴗ': u'ყ', u'ⴗ': u'ყ', u'Ⴘ': u'შ',
            u'ⴘ': u'შ', u'Ⴙ': u'ჩ', u'ⴙ': u'ჩ', u'Ⴚ': u'ც', u'ⴚ': u'ც',
            u'Ⴛ': u'ძ', u'ⴛ': u'ძ', u'Ⴜ': u'წ', u'ⴜ': u'წ', u'Ⴝ': u'ჭ',
            u'ⴝ': u'ჭ', u'Ⴞ': u'ხ', u'ⴞ': u'ხ', u'Ⴤ': u'ჴ', u'ⴤ': u'ჴ',
            u'Ⴟ': u'ჯ', u'ⴟ': u'ჯ', u'Ⴠ': u'ჰ', u'ⴠ': u'ჰ', u'Ⴥ': u'ჵ',
            u'ⴥ': u'ჵ'
        })


__lang__ = Georgian



from kss import split_sentences
from sentence_split import run_evaluate

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("datasets", nargs="+")
    parser.add_argument("--write_result")
    parser.add_argument("--write_err")
    parser.add_argument("--backend", default="mecab", choices=["mecab", "pecab", "punct", "fast"])
    args = parser.parse_args()

    split_sentences("foo-bar", backend=args.backend)  # warm-up

    for dataset in args.datasets:
        run_evaluate(
            dataset,
            lambda text: split_sentences(text, backend=args.backend),
            args.write_result,
            args.write_err,
        )



from kss import Kss


def test_split_sentences():
    split_sentences = Kss("split_sentences")
    text = "회사 동료 분들과 다녀왔는데 분위기도 좋고 음식도 맛있었어요 다만, 강남 토끼정이 강남 쉑쉑버거 골목길로 쭉 올라가야 하는데 다들 쉑쉑버거의 유혹에 넘어갈 뻔 했답니다 강남역 맛집 토끼정의 외부 모습."
    output = split_sentences(text)
    assert output == ['회사 동료 분들과 다녀왔는데 분위기도 좋고 음식도 맛있었어요',
                      '다만, 강남 토끼정이 강남 쉑쉑버거 골목길로 쭉 올라가야 하는데 다들 쉑쉑버거의 유혹에 넘어갈 뻔 했답니다',
                      '강남역 맛집 토끼정의 외부 모습.']



from functools import partial
from typing import Union, List, Tuple

import distance
from kss._modules.hangulization.hangulize import hangulize as _hangulize
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers

supported_langs = languages = {
    "lat": "라틴어",
    "deu": "독일어",
    "rus": "러시아어",
    "ell": "현대 그리스어",
    "grc": "고대 그리스어",
    "ita": "이탈리아어",
    "spa": "스페인어",
    "jpn": "일본어",
    "pol": "폴란드어",
    "ces": "체코어",
    "hbs": "세르보크로아트어",
    "ron": "루마니아어",
    "hun": "헝가리어",
    "vie": "베트남어",
    "swe": "스웨덴어",
    "nld": "네덜란드어",
    "por": "포르투갈어",
    "por.br": "브라질 포르투갈어",
    "cym": "웨일스어",
    "wlm": "중세 웨일스어",
    "bul": "불가리아어",
    "kat": "조지아어 간략전사",
    "kat.narrow": "조지아어 정밀전사",
    "fin": "핀란드어",
    "est": "에스토니아어",
    "lav": "라트비아어",
    "lit": "리투아니아어",
    "isl": "아이슬란드어",
    "cat": "카탈루냐어",
    "slk": "슬로바키아어",
    "slv": "슬로베니아어",
    "sqi": "알바니아어",
    "epo": "에스페란토",
    "ukr": "우크라이나어",
    "bel": "벨라루스어",
    "mkd": "마케도니아어",
    "tur": "터키어",
    "aze": "아제르바이잔어"
}


def hangulize(
    text: Union[str, List[str], Tuple[str]],
    lang: str,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts the given text to Hangul pronunciation.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        lang (str): source language code
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: Hangul pronunciation of the given text

    Examples:
        >>> from kss import Kss
        >>> hangulize = Kss("hangulize")
        >>> text = "Giro d'Italia"
        >>> output = hangulize(text, lang="ita")
        >>> print(output)
        "지로 디탈리아"

    References:
        This was copied from [hangulize](https://github.com/sublee/hangulize) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    lang = _check_lang(lang)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(_hangulize, code=lang),
        inputs=text,
        num_workers=num_workers,
    )


def _check_lang(lang: str):
    lang = lang.lower()
    lang = _check_type(lang, "lang", str)

    if lang not in supported_langs:
        error_message = f"'{lang}' is not supported language code." \
                        f"\nSupported language codes are the following:\n{list(supported_langs.items())}"
        closest_lang = _find_closest_lang(lang)
        if closest_lang:
            error_message += f"\n\nDid you mean '{closest_lang}'?"
        raise ValueError(error_message)

    return lang


def _find_closest_lang(lang, min_distance=0.5):
    current_min_distance = 99
    closest_lang = None
    for supported_lang in supported_langs:
        dist = distance.nlevenshtein(lang, supported_lang)
        if dist < current_min_distance and dist <= min_distance:
            current_min_distance = dist
            closest_lang = supported_lang

    if closest_lang is None:
        return None
    else:
        return closest_lang, supported_langs[closest_lang]



from kss import Kss


def test_morphemes():
    split_morphemes = Kss("split_morphemes")
    text = "아버지가방에들어오시다."
    output = split_morphemes(text)
    assert output == [('아버지', 'NNG'), ('가', 'JKS'), ('방', 'NNG'), ('에', 'JKB'), ('들어오', 'VV'), ('시', 'EP'), ('다', 'EF'), ('.', 'SF')]


if __name__ == '__main__':
    test_morphemes()



import sys
from functools import partial
from typing import List, Optional, Tuple, Union

from kss._modules.preprocessing.anonymize import _anonymize
from kss._modules.preprocessing.filter_out import _filter_out
from kss._modules.preprocessing.normalize import _normalize
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers


def preprocess(
    text: Union[str, List[str], Tuple[str]],
    normalization_type: Optional[str] = None,
    allow_doubled_spaces: bool = True,
    allow_html_tags: bool = True,
    allow_html_escape: bool = True,
    allow_halfwidth_hangul: bool = True,
    allow_hangul_jamo: bool = True,
    allow_invisible_chars: bool = True,
    reduce_char_repeats_over: int = sys.maxsize,
    reduce_emoticon_repeats_over: int = sys.maxsize,
    min_length: int = 0,
    max_length: int = sys.maxsize,
    min_mean_words_length: int = 0,
    max_mean_words_length: int = sys.maxsize,
    min_words: int = 0,
    max_words: int = sys.maxsize,
    min_lines: int = 0,
    max_lines: int = sys.maxsize,
    min_paragraphs: int = 0,
    max_paragraphs: int = sys.maxsize,
    min_alphabet_ratio: float = 0,
    max_alphabet_ratio: float = 1,
    min_alphanumeric_ratio: float = 0,
    max_alphanumeric_ratio: float = 1,
    min_number_ratio: float = 0,
    max_number_ratio: float = 1,
    min_punctuation_ratio: float = 0,
    max_punctuation_ratio: float = 1,
    min_symbols_to_words_ratio: float = 0,
    max_symbols_to_words_ratio: float = 0,
    min_lines_started_with_bullets_ratio: float = 0,
    max_lines_started_with_bullets_ratio: float = 1,
    min_whitespace_ratio: float = 0,
    max_whitespace_ratio: float = 1,
    min_parenthesis_ratio: float = 0,
    max_parenthesis_ratio: float = 1,
    min_ellipsis_ratio: float = 0,
    max_ellipsis_ratio: float = 1,
    min_hangul_ratio: float = 0,
    max_hangul_ratio: float = 1,
    max_hangul_incompleted_form_ratio: float = 1,
    max_words_length: int = sys.maxsize,
    max_line_repeats: int = sys.maxsize,
    max_line_by_char_repeats: int = sys.maxsize,
    max_paragraph_repeats: int = sys.maxsize,
    max_paragraph_by_char_repeats: int = sys.maxsize,
    max_repeating_top_ngram_repeats_score: float = sys.maxsize,
    max_repeating_duplicate_ngrams_score: float = sys.maxsize,
    ngram_size_for_repeating_top_ngram_repeats: int = 3,
    ngram_size_for_repeating_duplicate_ngrams: int = 3,
    phone_number_anonymization: bool = True,
    rrn_anonymization: bool = True,
    card_anonymization: bool = True,
    email_anonymization: bool = True,
    bank_account_anonymization: bool = True,
    credit_card_anonymization: bool = True,
    zip_anonymization: bool = True,
    bitcoin_anonymization: bool = True,
    url_anonymization: bool = True,
    ip_v6_anonymization: bool = True,
    ip_v4_anonymization: bool = True,
    phone_number_replacement: str = "<PHONE_NUMBER>",
    rrn_replacement: str = "<RRN>",
    card_replacement: str = "<CARD>",
    email_replacement: str = "<EMAIL>",
    bank_account_replacement: str = "<BANK_ACCOUNT>",
    credit_card_replacement: str = "<CREDIT_CARD>",
    zip_replacement: str = "<ZIP>",
    bitcoin_replacement: str = "<BITCOIN>",
    url_replacement: str = "<URL>",
    ip_v6_replacement: str = "<IPV6>",
    ip_v4_replacement: str = "<IPV4>",
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This preprocesses text with various options.
    This does 1) normalization, 2) filtering out, and 3) anonymization in order.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        normalization_type (Optional[str]): normalization type
        allow_doubled_spaces (bool): whether to allow doubled spaces or not
        allow_html_tags (bool): whether to allow HTML tags or not
        allow_html_escape (bool): whether to allow HTML escape or not
        allow_halfwidth_hangul (bool): whether to allow halfwidth Hangul or not
        allow_hangul_jamo (bool): whether to allow Hangul jamo or not
        allow_invisible_chars (bool): whether to allow invisible characters or not
        reduce_char_repeats_over (int): the maximum number of character that can be repeated
        reduce_emoticon_repeats_over (int): the maximum number of emoticon that can be repeated
        min_length (int): minimum length of text
        max_length (int): maximum length of text
        min_mean_words_length (int): minimum mean words length
        max_mean_words_length (int): maximum mean words length
        min_words (int): minimum number of words
        max_words (int): maximum number of words
        min_lines (int): minimum number of lines
        max_lines (int): maximum number of lines
        min_paragraphs (int): minimum number of paragraphs
        max_paragraphs (int): maximum number of paragraphs
        min_alphabet_ratio (float): minimum alphabet ratio
        max_alphabet_ratio (float): maximum alphabet ratio
        min_alphanumeric_ratio (float): minimum alphanumeric ratio
        max_alphanumeric_ratio (float): maximum alphanumeric ratio
        min_number_ratio (float): minimum number ratio
        max_number_ratio (float): maximum number ratio
        min_punctuation_ratio (float): minimum punctuation ratio
        max_punctuation_ratio (float): maximum punctuation ratio
        min_symbols_to_words_ratio (float): minimum symbols to words ratio
        max_symbols_to_words_ratio (float): maximum symbols to words ratio
        min_lines_started_with_bullets_ratio (float): minimum lines started with bullets ratio
        max_lines_started_with_bullets_ratio (float): maximum lines started with bullets ratio
        min_whitespace_ratio (float): minimum whitespace ratio
        max_whitespace_ratio (float): maximum whitespace ratio
        min_parenthesis_ratio (float): minimum parenthesis ratio
        max_parenthesis_ratio (float): maximum parenthesis ratio
        min_ellipsis_ratio (float): minimum ellipsis ratio
        max_ellipsis_ratio (float): maximum ellipsis ratio
        min_hangul_ratio (float): minimum Hangul ratio
        max_hangul_ratio (float): maximum Hangul ratio
        max_hangul_incompleted_form_ratio (float): maximum Hangul non-completed form ratio
        max_words_length (int): maximum words length
        max_line_repeats (int): maximum line repeats
        max_line_by_char_repeats (int): maximum line by char repeats
        max_paragraph_repeats (int): maximum paragraph repeats
        max_paragraph_by_char_repeats (int): maximum paragraph by char repeats
        max_repeating_top_ngram_repeats_score (float): maximum repeating top ngram repeats score
        max_repeating_duplicate_ngrams_score (float): maximum repeating duplicate ngrams score
        ngram_size_for_repeating_top_ngram_repeats (int): ngram size for repeating top ngram repeats
        ngram_size_for_repeating_duplicate_ngrams (int): ngram size for repeating duplicate ngrams
        phone_number_anonymization (bool): whether to anonymize phone number or not
        rrn_anonymization (bool): whether to anonymize RRN or not
        card_anonymization (bool): whether to anonymize card or not
        email_anonymization (bool): whether to anonymize email or not
        bank_account_anonymization (bool): whether to anonymize bank account or not
        credit_card_anonymization (bool): whether to anonymize credit card or not
        zip_anonymization (bool): whether to anonymize zip or not
        bitcoin_anonymization (bool): whether to anonymize bitcoin or not
        url_anonymization (bool): whether to anonymize URL or not
        ip_v6_anonymization (bool): whether to anonymize IPv6 or not
        ip_v4_anonymization (bool): whether to anonymize IPv4 or not
        phone_number_replacement (str): replacement for phone number
        rrn_replacement (str): replacement for RRN
        card_replacement (str): replacement for card
        email_replacement (str): replacement for email
        bank_account_replacement (str): replacement for bank account
        credit_card_replacement (str): replacement for credit card
        zip_replacement (str): replacement for zip
        bitcoin_replacement (str): replacement for bitcoin
        url_replacement (str): replacement for URL
        ip_v6_replacement (str): replacement for IPv6
        ip_v4_replacement (str): replacement for IPv4
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[Tuple[str, Dict[str, Any]], List[Tuple[str, Dict[str, Any]]]]:
            preprocessed text and filtering metadata or list of preprocessed texts and filtering metadata
    """

    text, finish = _check_text(text)

    if finish:
        return text

    if normalization_type is not None:
        normalization_type = _check_type(normalization_type, "normalization_type", str)
    allow_doubled_spaces = _check_type(allow_doubled_spaces, "allow_doubled_spaces", bool)
    allow_html_tags = _check_type(allow_html_tags, "allow_html_tags", bool)
    allow_html_escape = _check_type(allow_html_escape, "allow_html_escape", bool)
    allow_halfwidth_hangul = _check_type(allow_halfwidth_hangul, "allow_halfwidth_hangul", bool)
    allow_hangul_jamo = _check_type(allow_hangul_jamo, "allow_hangul_jamo", bool)
    reduce_char_repeats_over = _check_type(reduce_char_repeats_over, "reduce_char_repeats_over", int)
    reduce_emoticon_repeats_over = _check_type(reduce_emoticon_repeats_over, "reduce_emoticon_repeats_over", int)
    min_length = _check_type(min_length, "min_length", int)
    max_length = _check_type(max_length, "max_length", int)
    min_mean_words_length = _check_type(min_mean_words_length, "min_mean_words_length", int)
    max_mean_words_length = _check_type(max_mean_words_length, "max_mean_words_length", int)
    min_words = _check_type(min_words, "min_words", int)
    max_words = _check_type(max_words, "max_words", int)
    min_lines = _check_type(min_lines, "min_lines", int)
    max_lines = _check_type(max_lines, "max_lines", int)
    min_paragraphs = _check_type(min_paragraphs, "min_paragraphs", int)
    max_paragraphs = _check_type(max_paragraphs, "max_paragraphs", int)
    min_alphabet_ratio = _check_type(min_alphabet_ratio, "min_alphabet_ratio", float)
    max_alphabet_ratio = _check_type(max_alphabet_ratio, "max_alphabet_ratio", float)
    min_alphanumeric_ratio = _check_type(min_alphanumeric_ratio, "min_alphanumeric_ratio", float)
    max_alphanumeric_ratio = _check_type(max_alphanumeric_ratio, "max_alphanumeric_ratio", float)
    min_number_ratio = _check_type(min_number_ratio, "min_number_ratio", float)
    max_number_ratio = _check_type(max_number_ratio, "max_number_ratio", float)
    min_punctuation_ratio = _check_type(min_punctuation_ratio, "min_punctuation_ratio", float)
    max_punctuation_ratio = _check_type(max_punctuation_ratio, "max_punctuation_ratio", float)
    min_symbols_to_words_ratio = _check_type(min_symbols_to_words_ratio, "min_symbols_to_words_ratio", float)
    max_symbols_to_words_ratio = _check_type(max_symbols_to_words_ratio, "max_symbols_to_words_ratio", float)
    min_lines_started_with_bullets_ratio = _check_type(min_lines_started_with_bullets_ratio,
                                                       "min_lines_started_with_bullets_ratio", float)
    max_lines_started_with_bullets_ratio = _check_type(max_lines_started_with_bullets_ratio,
                                                       "max_lines_started_with_bullets_ratio", float)
    min_whitespace_ratio = _check_type(min_whitespace_ratio, "min_whitespace_ratio", float)
    max_whitespace_ratio = _check_type(max_whitespace_ratio, "max_whitespace_ratio", float)
    min_parenthesis_ratio = _check_type(min_parenthesis_ratio, "min_parenthesis_ratio", float)
    max_parenthesis_ratio = _check_type(max_parenthesis_ratio, "max_parenthesis_ratio", float)
    min_ellipsis_ratio = _check_type(min_ellipsis_ratio, "min_ellipsis_ratio", float)
    max_ellipsis_ratio = _check_type(max_ellipsis_ratio, "max_ellipsis_ratio", float)
    min_hangul_ratio = _check_type(min_hangul_ratio, "min_hangul_ratio", float)
    max_hangul_ratio = _check_type(max_hangul_ratio, "max_hangul_ratio", float)
    max_words_length = _check_type(max_words_length, "max_words_length", int)
    max_line_repeats = _check_type(max_line_repeats, "max_line_repeats", int)
    max_line_by_char_repeats = _check_type(max_line_by_char_repeats, "max_line_by_char_repeats", int)
    max_paragraph_repeats = _check_type(max_paragraph_repeats, "max_paragraph_repeats", int)
    max_paragraph_by_char_repeats = _check_type(max_paragraph_by_char_repeats, "max_paragraph_by_char_repeats", int)
    max_repeating_top_ngram_repeats_score = _check_type(max_repeating_top_ngram_repeats_score,
                                                        "max_repeating_top_ngram_repeats_score", float)
    max_repeating_duplicate_ngrams_score = _check_type(max_repeating_duplicate_ngrams_score,
                                                       "max_repeating_duplicate_ngrams_score", float)
    ngram_size_for_repeating_top_ngram_repeats = _check_type(ngram_size_for_repeating_top_ngram_repeats,
                                                             "ngram_size_for_repeating_top_ngram_repeats", int)
    ngram_size_for_repeating_duplicate_ngrams = _check_type(ngram_size_for_repeating_duplicate_ngrams,
                                                            "ngram_size_for_repeating_duplicate_ngrams", int)
    max_hangul_incompleted_form_ratio = _check_type(max_hangul_incompleted_form_ratio,
                                                      "max_hangul_incompleted_form_ratio", float)
    phone_number_anonymization = _check_type(phone_number_anonymization, "phone_number_anonymization", bool)
    rrn_anonymization = _check_type(rrn_anonymization, "rrn_anonymization", bool)
    card_anonymization = _check_type(card_anonymization, "card_anonymization", bool)
    email_anonymization = _check_type(email_anonymization, "email_anonymization", bool)
    bank_account_anonymization = _check_type(bank_account_anonymization, "bank_account_anonymization", bool)
    credit_card_anonymization = _check_type(credit_card_anonymization, "credit_card_anonymization", bool)
    zip_anonymization = _check_type(zip_anonymization, "zip_anonymization", bool)
    bitcoin_anonymization = _check_type(bitcoin_anonymization, "bitcoin_anonymization", bool)
    url_anonymization = _check_type(url_anonymization, "url_anonymization", bool)
    ip_v6_anonymization = _check_type(ip_v6_anonymization, "ip_v6_anonymization", bool)
    ip_v4_anonymization = _check_type(ip_v4_anonymization, "ip_v4_anonymization", bool)
    phone_number_replacement = _check_type(phone_number_replacement, "phone_number_replacement", str)
    rrn_replacement = _check_type(rrn_replacement, "rrn_replacement", str)
    card_replacement = _check_type(card_replacement, "card_replacement", str)
    email_replacement = _check_type(email_replacement, "email_replacement", str)
    bank_account_replacement = _check_type(bank_account_replacement, "bank_account_replacement", str)
    credit_card_replacement = _check_type(credit_card_replacement, "credit_card_replacement", str)
    zip_replacement = _check_type(zip_replacement, "zip_replacement", str)
    bitcoin_replacement = _check_type(bitcoin_replacement, "bitcoin_replacement", str)
    url_replacement = _check_type(url_replacement, "url_replacement", str)
    ip_v6_replacement = _check_type(ip_v6_replacement, "ip_v6_replacement", str)
    ip_v4_replacement = _check_type(ip_v4_replacement, "ip_v4_replacement", str)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(
            _preprocess,
            normalization_type=normalization_type,
            allow_doubled_spaces=allow_doubled_spaces,
            allow_html_tags=allow_html_tags,
            allow_html_escape=allow_html_escape,
            allow_halfwidth_hangul=allow_halfwidth_hangul,
            allow_hangul_jamo=allow_hangul_jamo,
            allow_invisible_chars=allow_invisible_chars,
            reduce_char_repeats_over=reduce_char_repeats_over,
            reduce_emoticon_repeats_over=reduce_emoticon_repeats_over,
            min_length=min_length,
            max_length=max_length,
            min_mean_words_length=min_mean_words_length,
            max_mean_words_length=max_mean_words_length,
            min_words=min_words,
            max_words=max_words,
            min_lines=min_lines,
            max_lines=max_lines,
            min_paragraphs=min_paragraphs,
            max_paragraphs=max_paragraphs,
            min_alphabet_ratio=min_alphabet_ratio,
            max_alphabet_ratio=max_alphabet_ratio,
            min_alphanumeric_ratio=min_alphanumeric_ratio,
            max_alphanumeric_ratio=max_alphanumeric_ratio,
            min_number_ratio=min_number_ratio,
            max_number_ratio=max_number_ratio,
            min_punctuation_ratio=min_punctuation_ratio,
            max_punctuation_ratio=max_punctuation_ratio,
            min_symbols_to_words_ratio=min_symbols_to_words_ratio,
            max_symbols_to_words_ratio=max_symbols_to_words_ratio,
            min_lines_started_with_bullets_ratio=min_lines_started_with_bullets_ratio,
            max_lines_started_with_bullets_ratio=max_lines_started_with_bullets_ratio,
            min_whitespace_ratio=min_whitespace_ratio,
            max_whitespace_ratio=max_whitespace_ratio,
            min_parenthesis_ratio=min_parenthesis_ratio,
            max_parenthesis_ratio=max_parenthesis_ratio,
            min_ellipsis_ratio=min_ellipsis_ratio,
            max_ellipsis_ratio=max_ellipsis_ratio,
            min_hangul_ratio=min_hangul_ratio,
            max_hangul_ratio=max_hangul_ratio,
            max_hangul_incompleted_form_ratio=max_hangul_incompleted_form_ratio,
            max_words_length=max_words_length,
            max_line_repeats=max_line_repeats,
            max_line_by_char_repeats=max_line_by_char_repeats,
            max_paragraph_repeats=max_paragraph_repeats,
            max_paragraph_by_char_repeats=max_paragraph_by_char_repeats,
            max_repeating_top_ngram_repeats_score=max_repeating_top_ngram_repeats_score,
            max_repeating_duplicate_ngrams_score=max_repeating_duplicate_ngrams_score,
            ngram_size_for_repeating_top_ngram_repeats=ngram_size_for_repeating_top_ngram_repeats,
            ngram_size_for_repeating_duplicate_ngrams=ngram_size_for_repeating_duplicate_ngrams,
            phone_number_anonymization=phone_number_anonymization,
            rrn_anonymization=rrn_anonymization,
            card_anonymization=card_anonymization,
            email_anonymization=email_anonymization,
            bank_account_anonymization=bank_account_anonymization,
            credit_card_anonymization=credit_card_anonymization,
            zip_anonymization=zip_anonymization,
            bitcoin_anonymization=bitcoin_anonymization,
            url_anonymization=url_anonymization,
            ip_v6_anonymization=ip_v6_anonymization,
            ip_v4_anonymization=ip_v4_anonymization,
            phone_number_replacement=phone_number_replacement,
            rrn_replacement=rrn_replacement,
            card_replacement=card_replacement,
            email_replacement=email_replacement,
            bank_account_replacement=bank_account_replacement,
            credit_card_replacement=credit_card_replacement,
            zip_replacement=zip_replacement,
            bitcoin_replacement=bitcoin_replacement,
            url_replacement=url_replacement,
            ip_v6_replacement=ip_v6_replacement,
            ip_v4_replacement=ip_v4_replacement,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _preprocess(
    text,
    normalization_type=None,
    allow_doubled_spaces=True,
    allow_html_tags=True,
    allow_html_escape=True,
    allow_halfwidth_hangul=True,
    allow_hangul_jamo=True,
    allow_invisible_chars=True,
    reduce_char_repeats_over=sys.maxsize,
    reduce_emoticon_repeats_over=sys.maxsize,
    min_length=0,
    max_length=sys.maxsize,
    min_mean_words_length=0,
    max_mean_words_length=sys.maxsize,
    min_words=0,
    max_words=sys.maxsize,
    min_lines=0,
    max_lines=sys.maxsize,
    min_paragraphs=0,
    max_paragraphs=sys.maxsize,
    min_alphabet_ratio=0,
    max_alphabet_ratio=1,
    min_alphanumeric_ratio=0,
    max_alphanumeric_ratio=1,
    min_number_ratio=0,
    max_number_ratio=1,
    min_punctuation_ratio=0,
    max_punctuation_ratio=1,
    min_symbols_to_words_ratio=0,
    max_symbols_to_words_ratio=0,
    min_lines_started_with_bullets_ratio=0,
    max_lines_started_with_bullets_ratio=1,
    min_whitespace_ratio=0,
    max_whitespace_ratio=1,
    min_parenthesis_ratio=0,
    max_parenthesis_ratio=1,
    min_ellipsis_ratio=0,
    max_ellipsis_ratio=1,
    min_hangul_ratio=0,
    max_hangul_ratio=1,
    max_hangul_incompleted_form_ratio=1,
    max_words_length=sys.maxsize,
    max_line_repeats=sys.maxsize,
    max_line_by_char_repeats=sys.maxsize,
    max_paragraph_repeats=sys.maxsize,
    max_paragraph_by_char_repeats=sys.maxsize,
    max_repeating_top_ngram_repeats_score=sys.maxsize,
    max_repeating_duplicate_ngrams_score=sys.maxsize,
    ngram_size_for_repeating_top_ngram_repeats=3,
    ngram_size_for_repeating_duplicate_ngrams=3,
    phone_number_anonymization=True,
    rrn_anonymization=True,
    card_anonymization=True,
    email_anonymization=True,
    bank_account_anonymization=True,
    credit_card_anonymization=True,
    zip_anonymization=True,
    bitcoin_anonymization=True,
    url_anonymization=True,
    ip_v6_anonymization=True,
    ip_v4_anonymization=True,
    phone_number_replacement="<PHONE_NUMBER>",
    rrn_replacement="<RRN>",
    card_replacement="<CARD>",
    email_replacement="<EMAIL>",
    bank_account_replacement="<BANK_ACCOUNT>",
    credit_card_replacement="<CREDIT_CARD>",
    zip_replacement="<ZIP>",
    bitcoin_replacement="<BITCOIN>",
    url_replacement="<URL>",
    ip_v6_replacement="<IPV6>",
    ip_v4_replacement="<IPV4>",
):
    # 1. normalize text
    text = _normalize(
        text,
        normalization_type=normalization_type,
        allow_doubled_spaces=allow_doubled_spaces,
        allow_html_tags=allow_html_tags,
        allow_html_escape=allow_html_escape,
        allow_halfwidth_hangul=allow_halfwidth_hangul,
        allow_hangul_jamo=allow_hangul_jamo,
        allow_invisible_chars=allow_invisible_chars,
        reduce_char_repeats_over=reduce_char_repeats_over,
        reduce_emoticon_repeats_over=reduce_emoticon_repeats_over,
    )

    # 2. filter out bad text
    is_filtered_out, filtered_reason = _filter_out(
        text,
        min_length=min_length,
        max_length=max_length,
        min_mean_words_length=min_mean_words_length,
        max_mean_words_length=max_mean_words_length,
        min_words=min_words,
        max_words=max_words,
        min_lines=min_lines,
        max_lines=max_lines,
        min_paragraphs=min_paragraphs,
        max_paragraphs=max_paragraphs,
        min_alphabet_ratio=min_alphabet_ratio,
        max_alphabet_ratio=max_alphabet_ratio,
        min_alphanumeric_ratio=min_alphanumeric_ratio,
        max_alphanumeric_ratio=max_alphanumeric_ratio,
        min_number_ratio=min_number_ratio,
        max_number_ratio=max_number_ratio,
        min_punctuation_ratio=min_punctuation_ratio,
        max_punctuation_ratio=max_punctuation_ratio,
        min_symbols_to_words_ratio=min_symbols_to_words_ratio,
        max_symbols_to_words_ratio=max_symbols_to_words_ratio,
        min_lines_started_with_bullets_ratio=min_lines_started_with_bullets_ratio,
        max_lines_started_with_bullets_ratio=max_lines_started_with_bullets_ratio,
        min_whitespace_ratio=min_whitespace_ratio,
        max_whitespace_ratio=max_whitespace_ratio,
        min_parenthesis_ratio=min_parenthesis_ratio,
        max_parenthesis_ratio=max_parenthesis_ratio,
        min_ellipsis_ratio=min_ellipsis_ratio,
        max_ellipsis_ratio=max_ellipsis_ratio,
        min_hangul_ratio=min_hangul_ratio,
        max_hangul_ratio=max_hangul_ratio,
        max_words_length=max_words_length,
        max_line_repeats=max_line_repeats,
        max_line_by_char_repeats=max_line_by_char_repeats,
        max_paragraph_repeats=max_paragraph_repeats,
        max_paragraph_by_char_repeats=max_paragraph_by_char_repeats,
        max_repeating_top_ngram_repeats_score=max_repeating_top_ngram_repeats_score,
        max_repeating_duplicate_ngrams_score=max_repeating_duplicate_ngrams_score,
        ngram_size_for_repeating_top_ngram_repeats=ngram_size_for_repeating_top_ngram_repeats,
        ngram_size_for_repeating_duplicate_ngrams=ngram_size_for_repeating_duplicate_ngrams,
        max_hangul_incompleted_form_ratio=max_hangul_incompleted_form_ratio,
    )

    if is_filtered_out:
        return None, filtered_reason

    # 3. anonymize text
    text = _anonymize(
        text,
        phone_number_anonymization=phone_number_anonymization,
        rrn_anonymization=rrn_anonymization,
        card_anonymization=card_anonymization,
        email_anonymization=email_anonymization,
        bank_account_anonymization=bank_account_anonymization,
        credit_card_anonymization=credit_card_anonymization,
        zip_anonymization=zip_anonymization,
        bitcoin_anonymization=bitcoin_anonymization,
        url_anonymization=url_anonymization,
        ip_v6_anonymization=ip_v6_anonymization,
        ip_v4_anonymization=ip_v4_anonymization,
        phone_number_replacement=phone_number_replacement,
        rrn_replacement=rrn_replacement,
        card_replacement=card_replacement,
        email_replacement=email_replacement,
        bank_account_replacement=bank_account_replacement,
        credit_card_replacement=credit_card_replacement,
        zip_replacement=zip_replacement,
        bitcoin_replacement=bitcoin_replacement,
        url_replacement=url_replacement,
        ip_v6_replacement=ip_v6_replacement,
        ip_v4_replacement=ip_v4_replacement,
    )

    return text, filtered_reason



from koalanlp import API
from koalanlp.Util import initialize, finalize
from koalanlp.proc import SentenceSplitter, Tagger

from sentence_split import run_evaluate

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("datasets", nargs="+")
    parser.add_argument("--write_result")
    parser.add_argument("--write_err")
    parser.add_argument(
        "--backend",
        default="OKT",
        choices=["OKT", "HNN", "KMR", "RHINO", "EUNJEON", "ARIRANG", "KKMA"],
    )
    args = parser.parse_args()

    initialize(**{args.backend: "LATEST"})
    if args.backend in ("OKT", "HNN"):
        splitter = SentenceSplitter(getattr(API, args.backend))
    else:
        tagger = Tagger(getattr(API, args.backend))
        splitter = lambda text: [sent.surfaceString() for sent in tagger(text)]

    splitter("foo-bar")  # warm-up

    for dataset in args.datasets:
        run_evaluate(dataset, splitter, args.write_result, args.write_err)

    finalize()



#
# This file is part of KoreanCodecs.
#
# Copyright(C) 2002-2003 Hye-Shik Chang <perky@FreeBSD.org>.
#
# KoreanCodecs is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# KoreanCodecs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with KoreanCodecs; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# $Id: hangul.py,v 1.2 2003/10/15 19:24:53 perky Exp $
#
from itertools import chain


class UnicodeHangulError(Exception):

    def __init__(self, msg):
        self.msg = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.msg

    __str__ = __repr__


Null = u''


class Jaeum:
    Codes = (u'\u3131', u'\u3132', u'\u3133', u'\u3134', u'\u3135', u'\u3136',
             #    G         GG          GS         N          NJ         NH
             u'\u3137', u'\u3138', u'\u3139', u'\u313a', u'\u313b', u'\u313c',
             #    D         DD          L          LG         LM         LB
             u'\u313d', u'\u313e', u'\u313f', u'\u3140', u'\u3141', u'\u3142',
             #    LS        LT          LP         LH         M          B
             u'\u3143', u'\u3144', u'\u3145', u'\u3146', u'\u3147', u'\u3148',
             #    BB        BS          S          SS         NG         J
             u'\u3149', u'\u314a', u'\u314b', u'\u314c', u'\u314d', u'\u314e')
    #    JJ        C           K          T          P          H
    Width = len(Codes)
    G, GG, GS, N, NJ, NH, D, DD, L, LG, LM, LB, LS, LT, LP, LH, M, B, \
        BB, BS, S, SS, NG, J, JJ, C, K, T, P, H = Codes
    Choseong = [G, GG, N, D, DD, L, M, B, BB, S, SS, NG, J, JJ, C, K, T, P, H]
    Jongseong = [Null, G, GG, GS, N, NJ, NH, D, L, LG, LM, LB, LS, LT, \
                 LP, LH, M, B, BS, S, SS, NG, J, C, K, T, P, H]
    MultiElement = {
        GG: (G, G), GS: (G, S), NJ: (N, J), NH: (N, H), DD: (D, D),
        LG: (L, G), LM: (L, M), LB: (L, B), LS: (L, S), LT: (L, T),
        LP: (L, P), LH: (L, H), BB: (B, B), BS: (B, S), SS: (S, S),
        JJ: (J, J)
    }


class Moeum:
    Codes = (u'\u314f', u'\u3150', u'\u3151', u'\u3152', u'\u3153', u'\u3154',
             #    A          AE        YA         YAE         EO         E
             u'\u3155', u'\u3156', u'\u3157', u'\u3158', u'\u3159', u'\u315a',
             #    YEO        YE        O          WA          WAE        OE
             u'\u315b', u'\u315c', u'\u315d', u'\u315e', u'\u315f', u'\u3160',
             #    YO         U         WEO        WE          WI         YU
             u'\u3161', u'\u3162', u'\u3163')
    #    EU         YI        I
    Width = len(Codes)
    A, AE, YA, YAE, EO, E, YEO, YE, O, WA, WAE, OE, YO, \
        U, WEO, WE, WI, YU, EU, YI, I = Codes
    Jungseong = list(Codes)
    MultiElement = {
        AE: (A, I), YAE: (YA, I), YE: (YEO, I), WA: (O, A), WAE: (O, A, I),
        OE: (O, I), WEO: (U, EO), WE: (U, E), WI: (U, I), YI: (EU, I)
    }


# Aliases for your convinience
Choseong = Jaeum.Choseong
Jungseong = Moeum.Jungseong
Jongseong = Jaeum.Jongseong

for name, code in chain(Jaeum.__dict__.items(), Moeum.__dict__.items()):
    if name.isupper() and len(name) <= 3:
        exec('%s = %s' % (name, repr(code)))
del name, code

# Unicode Hangul Syllables Characteristics
ZONE = (u'\uAC00', u'\uD7A3')
NCHOSEONG = len(Choseong)
NJUNGSEONG = len(Jungseong)
NJONGSEONG = len(Jongseong)
JBASE_CHOSEONG = u'\u1100'
JBASE_JUNGSEONG = u'\u1161'
JBASE_JONGSEONG = u'\u11A8'
CHOSEONG_FILLER = u'\u115F'
JUNGSEONG_FILLER = u'\u1160'

_ishangul = (
    lambda code:
    ZONE[0] <= code <= ZONE[1] or
    code in Jaeum.Codes or
    code in Moeum.Codes
)

# Alternative Suffixes : do not use outside
ALT_SUFFIXES = {
    u'\uc744': (u'\ub97c', u'\uc744'),  # reul, eul
    u'\ub97c': (u'\ub97c', u'\uc744'),  # reul, eul
    u'\uc740': (u'\ub294', u'\uc740'),  # neun, eun
    u'\ub294': (u'\ub294', u'\uc740'),  # neun, eun
    u'\uc774': (u'\uac00', u'\uc774'),  # yi, ga
    u'\uac00': (u'\uac00', u'\uc774'),  # yi, ga
    u'\uc640': (u'\uc640', u'\uacfc'),  # wa, gwa
    u'\uacfc': (u'\uc640', u'\uacfc'),  # wa, gwa
}

# Ida-Varitaion Suffixes : do not use outside
IDA_SUFFIXES = {
    u'(\uc774)': (u'', u'\uc774'),  # (yi)da
    u'(\uc785)': (17, u'\uc785'),  # (ip)nida
    u'(\uc778)': (4, u'\uc778'),  # (in)-
}


def isJaeum(u):
    if u:
        for c in u:
            if c not in Jaeum.Codes:
                break
        else:
            return True
    return False


def isMoeum(u):
    if u:
        for c in u:
            if c not in Moeum.Codes:
                break
        else:
            return True
    return False


def ishangul(u):
    if u:
        for c in u:
            if not _ishangul(c):
                break
        else:
            return True
    return False


def join(codes):
    """ Join function which makes hangul syllable from jamos """
    if len(codes) != 3:
        raise UnicodeHangulError("needs 3-element tuple")
    if not codes[0] or not codes[1]:  # single jamo
        return codes[0] or codes[1]

    return chr(
        0xac00 + (
            Choseong.index(codes[0]) * NJUNGSEONG +
            Jungseong.index(codes[1])
        ) * NJONGSEONG + Jongseong.index(codes[2])
    )


def split(code):
    """ Split function which splits hangul syllable into jamos """
    if len(code) != 1 or not _ishangul(code):
        raise UnicodeHangulError("needs 1 hangul letter")
    if code in Jaeum.Codes:
        return (code, Null, Null)
    if code in Moeum.Codes:
        return (Null, code, Null)

    code = ord(code) - 0xac00
    return (
        Choseong[int(code / (NJUNGSEONG * NJONGSEONG))],  # Python3000 safe
        Jungseong[int(code / NJONGSEONG) % NJUNGSEONG],
        Jongseong[code % NJONGSEONG]
    )


def conjoin(s):
    obuff = []
    ncur = 0

    while ncur < len(s):
        c = s[ncur]
        if JBASE_CHOSEONG <= c <= u'\u1112' or c == CHOSEONG_FILLER:  # starts with choseong
            if len(s) > ncur + 1 and JUNGSEONG_FILLER <= s[ncur + 1] <= u'\u1175':
                cho = Choseong[ord(c) - ord(JBASE_CHOSEONG)]
                jung = Jungseong[ord(s[ncur + 1]) - ord(JBASE_JUNGSEONG)]
                if len(s) > ncur + 2 and JBASE_JONGSEONG <= s[ncur + 2] <= u'\u11C2':
                    jong = Jongseong[ord(s[ncur + 2]) - ord(JBASE_JONGSEONG) + 1]
                    ncur += 2
                else:
                    jong = Null
                    ncur += 1
                obuff.append(join([cho, jung, jong]))
            else:
                obuff.append(join([Choseong[ord(c) - ord(JBASE_CHOSEONG)], Null, Null]))
        elif JBASE_JUNGSEONG <= c <= u'\u1175':
            obuff.append(join([Null, Jungseong[ord(c) - ord(JBASE_JUNGSEONG)], Null]))
        else:
            obuff.append(c)
        ncur += 1

    return u''.join(obuff)


def disjoint(s):
    obuff = []
    for c in s:
        if _ishangul(c):
            cho, jung, jong = split(c)
            if cho:
                obuff.append(chr(ord(JBASE_CHOSEONG) + Choseong.index(cho)))
            else:
                obuff.append(CHOSEONG_FILLER)

            if jung:
                obuff.append(chr(ord(JBASE_JUNGSEONG) + Jungseong.index(jung)))
            else:
                obuff.append(JUNGSEONG_FILLER)

            if jong:
                obuff.append(chr(ord(JBASE_JONGSEONG) + Jongseong.index(jong) - 1))
        else:
            obuff.append(c)
    return u''.join(obuff)


def _has_final(c):
    # for internal use only
    if u'\uac00' <= c <= u'\ud7a3':  # hangul
        return 1, (ord(c) - 0xac00) % 28 > 0
    else:
        return 0, c in u'013678.bklmnptLMNRZ'


# Iterator Emulator for ancient versions before 2.1
try:
    iter
except:
    class iter:
        def __init__(self, obj):
            self.obj = obj
            self.ptr = 0

        def next(self):
            try:
                return self.obj[self.ptr]
            finally:
                self.ptr += 1

# Nested scope lambda emulation for versions before 2.2
import sys

if sys.hexversion < 0x2020000:
    class plambda:
        def __init__(self, obj):
            self.obj = obj

        def __call__(self):
            return self.obj
else:
    plambda = None
del sys


def format(fmtstr, *args, **kwargs):
    if kwargs:
        argget = lambda: kwargs
        if plambda:
            argget = plambda(kwargs)
    else:
        argget = iter(args).next

    obuff = []
    ncur = escape = fmtinpth = 0
    ofmt = fmt = u''

    while ncur < len(fmtstr):
        c = fmtstr[ncur]

        if escape:
            obuff.append(c)
            escape = 0
            ofmt = u''
        elif c == u'\\':
            escape = 1
        elif fmt:
            fmt += c
            if not fmtinpth and c.isalpha():
                ofmt = fmt % argget()
                obuff.append(ofmt)
                fmt = u''
            elif fmtinpth and c == u')':
                fmtinpth = 0
            elif c == u'(':
                fmtinpth = 1
            elif c == u'%':
                obuff.append(u'%')
        elif c == u'%':
            fmt += c
            ofmt = u''
        else:
            if ofmt and ALT_SUFFIXES.has_key(c):
                obuff.append(ALT_SUFFIXES[c][
                                 _has_final(ofmt[-1])[1] and 1 or 0
                                 ])
            elif ofmt and IDA_SUFFIXES.has_key(fmtstr[ncur:ncur + 3]):
                sel = IDA_SUFFIXES[fmtstr[ncur:ncur + 3]]
                ishan, hasfinal = _has_final(ofmt[-1])

                if hasfinal:
                    obuff.append(sel[1])
                elif ishan:
                    if sel[0]:
                        obuff[-1] = obuff[-1][:-1] + chr(ord(ofmt[-1]) + sel[0])
                else:
                    obuff.append(sel[0] and sel[1])
                ncur += 2
            else:
                obuff.append(c)

            ofmt = u''

        ncur += 1

    return u''.join(obuff)



# This code was copied from Text Rank KR [https://github.com/theeluwin/textrankr]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

from functools import partial
from typing import List, Dict, Union, Tuple

from kss._modules.summarization.utils import (
    _parse_text_into_sentences,
    _build_sentence_graph,
    Sentence,
)
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import (
    _check_text,
    _check_num_workers,
    _check_analyzer_backend,
    _check_type,
    _check_value,
)


def summarize_sentences(
    text: Union[str, List[str], Tuple[str]],
    backend: str = "auto",
    num_workers: Union[str, int] = "auto",
    max_sentences: int = 3,
    tolerance: Union[float] = 0.05,
    strip: bool = True,
    ignores: List[str] = None,
) -> Union[List[str], List[List[str]]]:
    """
    This summarizes the given text, using TextRank algorithm.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported.
        num_workers (Union[int, str])): the number of multiprocessing workers
        max_sentences (int): the max number of sentences in a summarization result.
        tolerance (float): a threshold for omitting edge weights.
        strip (bool): strip all sentences or not
        ignores (List[str]): list of strings to ignore

    Returns:
        Union[List[str], List[List[str]]]: outputs of text summarization

    Examples:
        >>> from kss import Kss
        >>> summarize_sentences = Kss("summarize_sentences")
        >>> text = "개그맨 겸 가수 ‘개가수’ UV 유세윤이 신곡 발매 이후 많은 남편들의 응원을 받고 있다. 유세윤은 지난 3일 오후 6시 새 싱글 ‘마더 사커(Mother Soccer)(Feat. 수퍼비)’를 발매했다. ‘마더 사커’는 아내에 대한 서운한 마음을 위트 있고 강한 어조로 디스 하는 남편 유세윤의 마음을 담은 곡이다. 발매 후 소셜 미디어 상에서 화제를 모으고 있는 가운데, 가수 하동균은 “유세유니 괜찮겠어”라는 반응을 보이기도 했다. 누리꾼들은 ‘두 분의 원만한 합의가 있기를 바랍니다’, ‘집에는 들어갈 수 있겠나’ 등 유세윤의 귀가를 걱정하는 모습을 보였다. 유세윤은 점입가경으로 ‘마더 사커’ 챌린지를 시작, 자신의 SNS를 통해 “부부 싸움이 좀 커졌네요”라며 배우 송진우와 함께 촬영한 영상을 게재했다. 해당 영상에서는 양말을 신고 침대에 들어간 뒤 환호를 지르거나 화장실 불을 끄지 않고 도망가는 등 아내의 잔소리 유발 포인트를 살려 재치 있는 영상을 완성했다. 유세윤은 ‘마더 사커’를 통해 남편들의 마음을 대변해 주고 있는 한편 아내의 반응은 어떨지 궁금증을 모은다."
        >>> summarize_sentences(text)
        ['개그맨 겸 가수 ‘개가수’ UV 유세윤이 신곡 발매 이후 많은 남편들의 응원을 받고 있다.', '‘마더 사커’는 아내에 대한 서운한 마음을 위트 있고 강한 어조로 디스 하는 남편 유세윤의 마음을 담은 곡이다.', '유세윤은 ‘마더 사커’를 통해 남편들의 마음을 대변해 주고 있는 한편 아내의 반응은 어떨지 궁금증을 모은다.']

    References:
        This was copied from [textrankr](https://https://github.com/theeluwin/textrankr) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    _check_analyzer_backend(backend)
    num_workers = _check_num_workers(text, num_workers)
    max_sentences = _check_type(max_sentences, "max_sentences", int)
    max_sentences = _check_value(
        max_sentences, "max_sentences", lambda x: x > 0, "integer value in 1~N"
    )
    tolerance = _check_type(tolerance, "tolerance", float)

    return _run_job(
        func=partial(
            _summarize_sentences,
            backend=backend,
            max_sentences=max_sentences,
            tolerance=tolerance,
            strip=strip,
            ignores=ignores,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _summarize_sentences(
    text: str,
    backend: str = "auto",
    max_sentences: int = 3,
    tolerance: float = 0.05,
    strip: bool = True,
    ignores: List[str] = None,
) -> List[str]:
    """
    Summarizes the given text, using TextRank algorithm.

    Args:
        text (str): a single text to be summarized.
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported.
        max_sentences (int): the max number of sentences in a summarization result.
        tolerance (float): a threshold for omitting edge weights.
        strip (bool): strip all sentences or not
        ignores (List[str]): list of strings to ignore

    Returns:
        List[str]: summarized sentences
    """
    from networkx import pagerank

    # parse text
    sentences: List[Sentence] = _parse_text_into_sentences(text, backend, strip, ignores)

    # build graph
    graph = _build_sentence_graph(sentences, tolerance=tolerance)

    # run pagerank
    pageranks: Dict[Sentence, float] = pagerank(graph, weight="weight")

    # get top-k sentences
    sentences = sorted(pageranks, key=pageranks.get, reverse=True)
    sentences = sentences[:max_sentences]
    sentences = sorted(sentences, key=lambda sentence: sentence.index)

    # output summaries
    return [sentence.text for sentence in sentences]



from kss import Kss


def test_hangulize():
    hangulize = Kss("hangulize")
    text = "Giro d'Italia"
    output = hangulize(text, lang="ita")
    assert output == "지로 디탈리아"


if __name__ == '__main__':
    test_hangulize()



import re
import jamo


def _j2h_char(lead, vowel, tail=None):
    """Convert a jamo character to Hangul."""
    return jamo.j2h(lead, vowel, tail)


def _j2h(text, add_placeholder_for_leading_vowels=False):
    if add_placeholder_for_leading_vowels:
        # add 'ㅇ' for leading vowels (e.g. 'ㅐ플' -> '애플')
        # this is useful for g2p conversion
        text = re.sub(r"(^|[^\u1100-\u1112])([\u1161-\u1175])", r"\1ᄋ\2", text)
    # case 1: consonant + vowel + consonant
    syllables = set(re.findall(r"[\u1100-\u1112][\u1161-\u1175][\u11A8-\u11C2]", text))
    for syl in syllables:
        text = text.replace(syl, _j2h_char(*syl))
    # case 2: consonant + vowel
    syllables = set(re.findall(r"[\u1100-\u1112][\u1161-\u1175]", text))
    for syl in syllables:
        text = text.replace(syl, _j2h_char(*syl))
    # case 3: remained a single consonant or vowel
    return jamo.j2hcj(text)



# Core algorithm was copied from Kiwi (https://github.com/bab2min/kiwipiepy).
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]
import re
from functools import partial
from typing import Union, List, Tuple

from kss._elements.subclasses import Token
from kss._modules.morphemes.analyzers import Analyzer
from kss._modules.morphemes.split_morphemes import split_morphemes
from kss._modules.morphemes.utils import _reset_spaces
from kss._modules.sentences.split_sentences import _split_sentences
from kss._modules.spacing.utils import postprocess, postprocess_heuristic
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_backend_mecab_pecab_only, _check_num_workers

any_ws = re.compile(r"\s+")
space_insertable = r"(([^SUWX]|X[RS]|S[EH]).* ([NMI]|V[VAX]|VCN|XR|XPN|S[WLHN]))|(SN ([MI]|N[PR]|NN[GP]|V[VAX]|VCN|XR|XPN|S[WHN]))|((S[FPL]).* ([NMI]|V[VAX]|VCN|XR|XPN|S[WHN]))"
space_insertable = re.compile(space_insertable)

backup_dict = {
    "ㆍ": "/ㆍ/",
    "\n": "\u2424",  # Symbol for newline
    "\t": "\u2409",  # Symbol for horizontal tab
    "\r": "\u240D",  # Symbol for carriage return
    "\f": "\u240C",  # Symbol for form feed
    "\v": "\u240B",  # Symbol for vertical tab
}
restore_dict = {
    v: k for k, v in backup_dict.items()
}


def correct_spacing(
    text: Union[str, List[str], Tuple[str]],
    backend: str = "auto",
    num_workers: Union[int, str] = "auto",
    reset_whitespaces: bool = False,
    return_morphemes: bool = False,
) -> Union[str, List[str]]:
    """
    This corrects the spacing of the text.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list/tuple of texts
        backend (str): morpheme analyzer backend. 'mecab', 'pecab', 'punct' are supported
        num_workers (Union[int, str])): the number of multiprocessing workers
        reset_whitespaces (bool): reset whitespaces or not
        return_morphemes (bool): whether to return morphemes or not

    Returns:
        Union[str, List[str]]: corrected text or list of corrected texts

    Examples:
        >>> from kss import Kss
        >>> correct_spacing = Kss("correct_spacing")
        >>> text = "아버지가방에들어가시다"
        >>> correct_spacing(text)
        '아버지가 방에 들어가시다'

    References:
        This was copied from [Kiwi](https://github.com/bab2min/kiwipiepy) and [ko-prfrdr](https://github.com/ychoi-kr/ko-prfrdr)
        and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    backend_string = backend
    backend = _check_backend_mecab_pecab_only(backend)
    _num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(_correct_spacing,
                     backend=backend,
                     backend_string=backend_string,
                     reset_whitespaces=reset_whitespaces,
                     return_morphemes=return_morphemes),
        inputs=text,
        num_workers=_num_workers,
    )


def _correct_spacing(
    text: str,
    backend: Analyzer,
    backend_string: str,
    reset_whitespaces: bool = False,
    return_morphemes: bool = False,
) -> Union[str, Tuple[str, List[Tuple[str, str]]]]:
    if reset_whitespaces:
        text = text.replace(" ", "")

    for k, v in backup_dict.items():
        text = text.replace(k, v)

    tokens = []
    chunks = []
    char_length = 0
    last_position = 0
    prev_token = None

    morphs = split_morphemes(
        text,
        backend=backend_string,
        drop_space=False
    )

    for i, (word, pos) in enumerate(morphs):
        token = Token(
            text=word,
            pos=pos.split("+")[0],
            idx=i,
            start=char_length,
            end=char_length + len(word),
        )
        char_length += len(word)
        tokens.append(token)

    for t in tokens:
        if last_position < t.start:
            if (t.pos.startswith("E") or t.pos.startswith("J") or t.pos.startswith("XS")
                or t.pos == "VX" and t.text in "하지"
            ):
                s = any_ws.sub("", text[last_position:t.start])
            else:
                s = text[last_position:t.start]
            if s:
                chunks.append(s)
            last_position = t.start

        if prev_token and space_insertable.match(prev_token.pos + " " + t.pos):
            if t.pos == "VX" and t.text in "하지":
                pass
            elif len(chunks) != 0 and not chunks[-1][-1].isspace():
                chunks.append(" ")

        if last_position < t.end:
            s = any_ws.sub("", text[last_position:t.end])
            if s:
                chunks.append(s)

        last_position = t.end
        prev_token = t

    if last_position < len(text):
        chunks.append(text[last_position:])

    output_text = "".join(chunks)
    sents, morphs = _split_sentences(
        output_text,
        backend=backend,
        recursion=99,
        strip=True,
        return_morphemes=True
    )

    output_text = " ".join(sents)
    morphs_with_spaces = _reset_spaces(output_text, morphs)

    output_text = postprocess(output_text, morphs_with_spaces)
    output_text = postprocess_heuristic(output_text)

    for k, v in restore_dict.items():
        output_text = output_text.replace(k, v)

    if return_morphemes:
        morphs_with_spaces = _reset_spaces(output_text, morphs_with_spaces)
        return output_text, morphs_with_spaces
    else:
        return output_text



# -*- coding: utf-8 -*-
"""
    hangulize
    ~~~~~~~~~

    Korean Alphabet Transcription.

    :copyright: (c) 2010-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
"""
from __future__ import absolute_import
import importlib


__version__ = '0.0.9'
__all__ = ['hangulize', 'get_lang', 'supports']


def hangulize(string, code=None):
    """Transcribes a loanword to Hangul.

        >>> print hangulize(u'gloria', 'ita')
        글로리아

    :param string: a loan word
    :param code: a language code as ISO 639-3.
    """
    return get_lang(code).hangulize(string)


def get_lang(code):
    """Returns a language instance from the given code."""
    code_head = code.split('.', 1)[0]
    if len(code_head) < 2 or len(code_head) > 3:
        raise ValueError('%r is an invalid language code' % code)
    def make_lang(code, submods):
        try:
            code = '.'.join([code] + list(submods))
            return import_lang_module(code).__lang__()
        except ImportError:
            raise ValueError('Hangulize does not support %s' % code)
    # split module path
    if '.' in code:
        code = code.split('.')
        submods = code[1:]
        code = code[0]
    else:
        submods = ()
    return make_lang(code, submods)


def import_lang_module(code):
    """Imports a module from the given code."""
    return importlib.import_module(f'kss._modules.hangulization.hangulize.langs.{code}')


def supports(code):
    """Checks if hangulize supports the given language.

        >>> supports('ita')
        True
        >>> supports('kat.narrow')
        True
        >>> supports('kor')
        False
    """
    try:
        import_lang_module(code)
        return True
    except ImportError:
        return False


def supported(code):
    """Deprecated with 0.0.6. Use :func:`hangulize.supports` instead."""
    import warnings
    warnings.warn('supported() has been deprecated, use supports() instead',
                  DeprecationWarning)
    return supports(code)


# include all submodules.
for name in ['.models', '.normalization', '.processing']:
    module = importlib.import_module(name, __name__)
    __all__.extend(module.__all__)
    for attr in module.__all__:
        locals()[attr] = getattr(module, attr)



from functools import partial
from typing import List, Union, Tuple

from kss._modules.hanja.utils import translate
from kss._modules.hanja.utils import split_hanja as _split_hanja
from kss._modules.hanja.utils import is_hanja as _is_hanja
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_num_workers, _check_char, _check_type


def split_hanja(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[List[str], List[List[str]]]:
    """
    This splits the given text into hanja string and non-hanja string.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: hanja string and non-hanja string of the given text

    Examples:
        >>> from kss import Kss
        >>> split_hanja = Kss("split_hanja")
        >>> text = "大韓民國은 民主共和國이다."
        >>> output = split_hanja(text)
        >>> print(output)
        ["大韓民國", "은 ", "民主共和國", "이다."]

    References:
        This was copied from [hanja](https://github.com/suminb/hanja) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=_split_hanja,
        inputs=text,
        num_workers=num_workers,
    )


def is_hanja(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if the given character is a hanja character.

    Args:
        text (Union[str, List[str], Tuple[str]): single character or list of characters
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a hanja character or not

    Examples:
        >>> from kss import Kss
        >>> is_hanja = Kss("is_hanja")
        >>> text = "大"
        >>> output = is_hanja(text)
        >>> print(output)
        True
        >>> text = "대"
        >>> output = is_hanja(text)
        >>> print(output)
        False

    References:
        This was copied from [hanja](https://github.com/suminb/hanja) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=_is_hanja,
        inputs=text,
        num_workers=num_workers,
    )


def hanja2hangul(
    text: Union[str, List[str], Tuple[str]],
    combination: bool = False,
    reverse: bool = False,
    html: bool = False,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts hanja to hangul.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        combination (bool): whether to return hanja and hangul together or not
        reverse (bool): whether to reverse the order of hanja and hangul or not
        html (bool): whether to return html format or not
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: hanja to hangul converted text or list of texts

    Examples:
        >>> from kss import Kss
        >>> hanja2hangul = Kss("hanja2hangul")
        >>> text = "大韓民國은 民主共和國이다."
        >>> output = hanja2hangul(text)
        >>> print(output)
        "대한민국은 민주공화국이다."

    References:
        This was copied from [hanja](https://github.com/suminb/hanja) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)
    combination = _check_type(combination, "combination", bool)
    reverse = _check_type(reverse, "reverse", bool)
    html = _check_type(html, "html", bool)

    if combination is False and reverse is True:
        raise ValueError("`reverse` is only available when `combination` is True")
    if combination is False and html is True:
        raise ValueError("`html` is only available when `combination` is True")

    return _run_job(
        func=partial(
            _hanja2hangul,
            combination=combination,
            reverse=reverse,
            html=html,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _hanja2hangul(text, combination=False, reverse=False, html=False):
    # translate hangul to hanja
    if combination is False:
        return translate(text, 'substitution')
    elif combination is True and reverse is False and html is False:
        return translate(text, 'combination-text')
    elif combination is True and reverse is True and html is False:
        return translate(text, 'combination-text-reversed')
    elif combination is True and reverse is False and html is True:
        return translate(text, 'combination-html')
    elif combination is True and reverse is True and html is True:
        return translate(text, 'combination-html-reversed')
    else:
        raise ValueError("Invalid combination of `combination`, `reverse`, and `html`")



from functools import partial
from typing import Union, List, Tuple

from kss._modules.safety.utils import bad_words, exceptions, pattern
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_num_workers, _check_type


def is_unsafe(
    text: Union[str, List[str], Tuple[str]],
    return_matches: bool = False,
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool], List[bool], List[List[str]]]:
    """
    This checks if the text is unsafe or not.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        return_matches (bool): whether to return matches or not
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool], List[bool], List[List[str]]]:
            whether the text is unsafe or not
            or list of whether the texts are unsafe or not
            or list of matched bad words in the texts

    Examples:
        >>> from kss import Kss
        >>> is_unsafe = Kss("is_unsafe")
        >>> text = "안녕하세요"
        >>> is_unsafe(text)
        False
        >>> text = "안녕하세요. 씨발"
        >>> is_unsafe(text)
        True
        >>> text = ["안녕하세요", "안녕하세요. 씨발"]
        >>> is_unsafe(text)
        [False, True]
        >>> text = "안녕하세요. 씨발"
        >>> is_unsafe(text, return_matches=True)
        ['씨발']
        >>> text = ["안녕하세요", "안녕하세요. 씨발"]
        >>> is_unsafe(text, return_matches=True)
        [[], ['씨발']]
    """
    text, finish = _check_text(text)
    return_matches = _check_type(return_matches, "return_matches", bool)
    num_workers = _check_num_workers(text, num_workers)

    if finish:
        return [] if return_matches else False

    return _run_job(
        func=partial(_is_unsafe, return_matches=return_matches),
        inputs=text,
        num_workers=num_workers,
    )


def _is_unsafe(text: str, return_matches: bool = False):
    if return_matches:
        matches = _is_unsafe_regex(text, return_matches=True) + _is_unsafe_dict(text, return_matches=True)
        deduplicated_matches = []
        for match in matches:
            if match not in deduplicated_matches:
                deduplicated_matches.append(match)
        return deduplicated_matches
    else:
        return _is_unsafe_regex(text, return_matches=False) or _is_unsafe_dict(text, return_matches=False)


def _is_unsafe_regex(text: str, return_matches: bool = False):
    longest_matches = []
    for match in pattern.findall(text):
        longest = ""
        for m in match:
            if len(m) > len(longest):
                longest = m
        if len(longest) > 0:
            longest_matches.append(longest)

    if return_matches:
        return longest_matches
    else:
        return len(longest_matches) > 0


def _is_unsafe_dict(text: str, return_matches: bool = False):
    matches = []
    for word in bad_words:
        if word in text:
            found_bad_words = True
            if word in exceptions:
                for exception in exceptions[word]:
                    if exception in text:
                        found_bad_words = False
                        break
            if found_bad_words:
                if return_matches:
                    matches.append(word)
                else:
                    return True
    if return_matches:
        return matches
    else:
        return False



from functools import partial
from typing import Union, List, Tuple

from kss._modules.augmentation.replacement import SynonymReplacement
from kss._modules.augmentation.utils import correct_josa
from kss._utils.logger import highlight_diffs, logger
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers, _check_backend_mecab_pecab_only


def augment(
    text: Union[str, List[str], Tuple[str]],
    replacement_ratio: float = 0.3,
    josa_correction: bool = True,
    num_workers: Union[int, str] = "auto",
    backend: str = "auto",
    verbose: bool = False,
) -> Union[str, List[str]]:
    """
    Augments text with synonym replacement method and, 
    optionally it postprocesses the text by correcting josa.
    For this, Kss uses the Korean wordnet from KAIST.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        replacement_ratio (float): ratio of words to be replaced
        josa_correction (bool): whether to normalize josa or not
        num_workers (Union[int, str]): the number of multiprocessing workers
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported
        verbose (bool): whether to print verbose outputs or not

    Returns:
        Union[str, List[str]]: augmented text or list of augmented texts

    Examples:
        >>> from kss import Kss
        >>> augment = Kss("augment")
        >>> text = "앞서 지난해 11월, 보이저 1호는 명령을 수신하고 수행하는 데엔 문제가 없었지만 통신 장치에 문제가 생겨 과학·엔지니어링 데이터가 지구로 전송되지 않았던 바 있다. 당시 그들은 컴퓨터 시스템을 재시작하고 문제의 근본적인 원인을 파악하기 위해 명령을 보내려고 시도했고, 이달 1일 '포크'라는 명령을 보냈다."
        >>> output = augment(text)
        >>> print(output)
        "앞서 지난해 11월, 보이저 1호는 명령을 수신하고 시행하는 데엔 문제가 없었지만 송신 장비에 문제가 생겨 과학·엔지니어링 데이터가 지구로 전송되지 않았던 바 있다. 당시 그들은 컴퓨터 시스템을 재시작하고 문제의 근본적인 원인을 파악하기 위해 명령을 보내려고 시도했고, 이달 1일 '포크'라는 명령을 보냈다."

    References:
        This was copied from [KoEDA](https://github.com/toriving/KoEDA) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    replacement_ratio = _check_type(replacement_ratio, "replacement_ratio", float)
    josa_correction = _check_type(josa_correction, "josa_correction", bool)
    verbose = _check_type(verbose, "verbose", bool)
    num_workers = _check_num_workers(text, num_workers)
    _check_backend_mecab_pecab_only(backend)

    if num_workers is not False and verbose:
        verbose = False
        logger.warn(
            "Verbose mode is not supported for multiprocessing. "
            "It will be turned off automatically."
        )

    return _run_job(
        func=partial(
            _augment,
            replacement_ratio=replacement_ratio,
            josa_correction=josa_correction,
            backend=backend,
            verbose=verbose,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _augment(
    text: str,
    replacement_ratio: float = 0.3,
    josa_correction: bool = True,
    backend: str = "auto",
    verbose: bool = False,
):
    orig_text = text
    replacement = SynonymReplacement(backend=backend)  # I want WSD...
    text = replacement(text, p=replacement_ratio, verbose=verbose)

    if josa_correction:
        text = correct_josa(text)

    if verbose:
        print()
        print(highlight_diffs(orig_text, text).replace("\n", "\\n"))

    return text



from kss import Kss


def test_is_unsafe():
    is_unsafe = Kss("is_unsafe")
    text = "안녕하세요"
    assert is_unsafe(text) is False
    text = "안녕하세요. 씨발"
    assert is_unsafe(text) is True
    text = "안녕하세요. 씨발"
    assert is_unsafe(text, return_matches=True) == ['씨발']



from functools import partial
from typing import Union, List, Tuple

import hangul_jamo
import jamo

from kss._modules.jamo.utils import _j2h
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_num_workers, _check_text, _check_type, _check_char


def h2j(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of Hangul to jamo.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: jamo string of the given text

    Examples:
        >>> from kss import Kss
        >>> h2j = Kss("h2j")
        >>> text = "안녕하세요"
        >>> output = h2j(text)
        >>> print(output)
        '안녕하세요'

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.h2j,
        inputs=text,
        num_workers=num_workers,
    )


def h2hcj(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of Hangul to Hangul Compatibility Jamo.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: Hangul Compatibility Jamo string of the given text

    Examples:
        >>> from kss import Kss
        >>> h2hcj = Kss("h2hcj")
        >>> text = "안녕하세요"
        >>> output = h2hcj(text)
        >>> print(output)
        'ㅇㅏㄴㄴㅕㅇㅎㅏㅅㅔㅇㅛ'

    References:
        This was copied from [hangul-jamo](https://github.com/jonghwanhyeon/hangul-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=hangul_jamo.decompose,
        inputs=text,
        num_workers=num_workers,
    )


def j2h(
    text: Union[str, List[str], Tuple[str]],
    add_placeholder_for_leading_vowels: bool = False,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of jamo to Hangul.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        add_placeholder_for_leading_vowels (bool): add 'ㅇ' for leading vowels (e.g. 'ㅐ플' -> '애플')
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: Hangul string of the given text

    Examples:
        >>> from kss import Kss
        >>> j2h = Kss("j2h")
        >>> text = '안녕하세요'
        >>> output = j2h(text)
        >>> print(output)
        '안녕하세요'

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)
    add_placeholder_for_leading_vowels = _check_type(
        add_placeholder_for_leading_vowels, "add_placeholder_for_leading_vowels", bool
    )

    return _run_job(
        func=partial(_j2h, add_placeholder_for_leading_vowels=add_placeholder_for_leading_vowels),
        inputs=text,
        num_workers=num_workers,
    )


def j2hcj(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of jamo to Hangul Compatibility Jamo.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: Hangul Compatibility Jamo string of the given text

    Examples:
        >>> from kss import Kss
        >>> j2hcj = Kss("j2hcj")
        >>> text = '안녕하세요'
        >>> output = j2hcj(text)
        >>> print(output)
        'ㅇㅏㄴㄴㅕㅇㅎㅏㅅㅔㅇㅛ'

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.j2hcj,
        inputs=text,
        num_workers=num_workers,
    )


def hcj2h(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of Hangul Compatibility Jamo to Hangul.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: Hangul string of the given text

    Examples:
        >>> from kss import Kss
        >>> hcj2h = Kss("hcj2h")
        >>> text = 'ㅇㅏㄴㄴㅕㅇㅎㅏㅅㅔㅇㅛ'
        >>> output = hcj2h(text)
        >>> print(output)
        '안녕하세요'

    References:
        This was copied from [hangul-jamo](https://github.com/jonghwanhyeon/hangul-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=hangul_jamo.compose,
        inputs=text,
        num_workers=num_workers,
    )


def hcj2j(
    text: Union[str, List[str], Tuple[str]],
    position: str = "vowel",
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts a string of Hangul Compatibility Jamo to jamo.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        position (str): the position of the HCJ character to convert to jamo character, one of 'lead', 'vowel', 'tail'
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: jamo string of the given text

    Examples:
        >>> from kss import Kss
        >>> hcj2j = Kss("hcj2j")
        >>> text = 'ㅇ'
        >>> output = hcj2j(text)
        >>> print(output)
        'ᄋ'

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)
    position = _check_type(position, "position", str).lower()

    if position not in ["lead", "vowel", "tail"]:
        raise ValueError("position should be one of 'lead', 'vowel', 'tail'")

    return _run_job(
        func=partial(jamo.hcj2j, position=position),
        inputs=text,
        num_workers=num_workers,
    )


def is_jamo(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if a character is a jamo character.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a jamo character or not

    Examples:
        >>> from kss import Kss
        >>> is_jamo = Kss("is_jamo")
        >>> text = 'ᄋ'
        >>> output = is_jamo(text)
        >>> print(output)
        True

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.is_jamo,
        inputs=text,
        num_workers=num_workers,
    )


def is_jamo_modern(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if a character is a modern jamo character.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a modern jamo character or not

    Examples:
        >>> from kss import Kss
        >>> is_jamo_modern = Kss("is_jamo_modern")
        >>> text = 'ᄋ'
        >>> output = is_jamo_modern(text)
        >>> print(output)
        True

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.is_jamo_modern,
        inputs=text,
        num_workers=num_workers,
    )


def is_hcj(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if a character is a Hangul Compatibility Jamo character.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a Hangul Compatibility Jamo character or not

    Examples:
        >>> from kss import Kss
        >>> is_hcj = Kss("is_hcj")
        >>> text = 'ㅇ'
        >>> output = is_hcj(text)
        >>> print(output)
        True

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.is_hcj,
        inputs=text,
        num_workers=num_workers,
    )


def is_hcj_modern(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if a character is a modern Hangul Compatibility Jamo character.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a modern Hangul Compatibility Jamo character or not

    Examples:
        >>> from kss import Kss
        >>> is_hcj_modern = Kss("is_hcj_modern")
        >>> text = 'ㅇ'
        >>> output = is_hcj_modern(text)
        >>> print(output)
        True

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.is_hcj_modern,
        inputs=text,
        num_workers=num_workers,
    )


def is_hangul_char(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[bool, List[bool]]:
    """
    This checks if a character is a Hangul character.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[bool, List[bool]]: whether the given character is a Hangul character or not

    Examples:
        >>> from kss import Kss
        >>> is_hangul_char = Kss("is_hangul_char")
        >>> text = '안'
        >>> output = is_hangul_char(text)
        >>> print(output)
        True

    References:
        This was copied from [jamo](https://github.com/JDongian/python-jamo) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return False

    text = _check_char(text)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=jamo.is_hangul_char,
        inputs=text,
        num_workers=num_workers,
    )



# -*- coding: utf-8 -*-
"""
    hangulize.normalization
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2010-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
"""
import unicodedata


__all__ = ['normalize_roman']


def normalize_roman(string, additional=None):
    """Removes diacritics from the string and converts to lowercase::

        >>> normalize_roman(u'Eèé')
        u'eee'

    """
    if additional:
        safe = list(additional.keys()) + list(additional.values())
        def gen():
            for c in string:
                if c not in safe:
                    yield normalize_roman(c)
                elif c in additional:
                    yield additional[c]
                else:
                    yield c
        return ''.join(gen())
    chars = []
    for c in string:
        if unicodedata.category(c) == 'Lo':
            chars.append(c)
        else:
            nor = unicodedata.normalize('NFD', c)
            chars.extend(x for x in nor if unicodedata.category(x) != 'Mn')
    return ''.join(chars).lower()



from kiwipiepy import Kiwi
from sentence_split import run_evaluate

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("datasets", nargs="+")
    parser.add_argument("--write_result")
    parser.add_argument("--write_err")
    args = parser.parse_args()

    kiwi = Kiwi()
    kiwi.split_into_sents("foo-bar")  # warm-up

    for dataset in args.datasets:
        run_evaluate(
            dataset,
            lambda text: [s.text for s in kiwi.split_into_sents(text)],
            args.write_result,
            args.write_err,
        )



import hangul_jamo
import jamo

from kss import Kss


def test_h2j():
    h2j = Kss("h2j")
    text = "한글"
    output = h2j(text)
    assert output == jamo.h2j(text)


def test_h2hcj():
    h2hcj = Kss("h2hcj")
    text = "한글"
    output = h2hcj(text)
    assert output == hangul_jamo.decompose(text)


def test_j2h():
    j2h = Kss("j2h")
    text = jamo.h2j("한글")
    output = j2h(text)
    assert output == "한글"


def test_j2hcj():
    j2hcj = Kss("j2hcj")
    text = jamo.h2j("한글")
    output = j2hcj(text)
    assert output == hangul_jamo.decompose("한글")


def test_hcj2j():
    hcj2j = Kss("hcj2j")
    text = "ㅎ"
    output = hcj2j(text)
    assert output == jamo.h2j("ㅎ")


def test_hcj2h():
    hcj2h = Kss("hcj2h")
    text = hangul_jamo.decompose("한글")
    output = hcj2h(text)
    assert output == "한글"


if __name__ == '__main__':
    test_h2j()
    test_h2hcj()
    test_j2h()
    test_j2hcj()
    test_hcj2j()
    test_hcj2h()



# This code was copied from g2pk [https://github.com/kyubyong/g2pK]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

from kss._modules.g2p.utils import rule_id2text, gloss


def link1(inp, descriptive=False, verbose=False):
    rule = rule_id2text["13"]
    out = inp

    pairs = [("ᆨᄋ", "ᄀ"),
             ("ᆩᄋ", "ᄁ"),
             ("ᆫᄋ", "ᄂ"),
             ("ᆮᄋ", "ᄃ"),
             ("ᆯᄋ", "ᄅ"),
             ("ᆷᄋ", "ᄆ"),
             ("ᆸᄋ", "ᄇ"),
             ("ᆺᄋ", "ᄉ"),
             ("ᆻᄋ", "ᄊ"),
             ("ᆽᄋ", "ᄌ"),
             ("ᆾᄋ", "ᄎ"),
             ("ᆿᄋ", "ᄏ"),
             ("ᇀᄋ", "ᄐ"),
             ("ᇁᄋ", "ᄑ")]
    for str1, str2 in pairs:
        out = out.replace(str1, str2)

    gloss(verbose, out, inp, rule)
    return out


def link2(inp, descriptive=False, verbose=False):
    rule = rule_id2text["14"]
    out = inp

    pairs = [("ᆪᄋ", "ᆨᄊ"),
             ("ᆬᄋ", "ᆫᄌ"),
             ("ᆰᄋ", "ᆯᄀ"),
             ("ᆱᄋ", "ᆯᄆ"),
             ("ᆲᄋ", "ᆯᄇ"),
             ("ᆳᄋ", "ᆯᄊ"),
             ("ᆴᄋ", "ᆯᄐ"),
             ("ᆵᄋ", "ᆯᄑ"),
             ("ᆹᄋ", "ᆸᄊ")]
    for str1, str2 in pairs:
        out = out.replace(str1, str2)

    gloss(verbose, out, inp, rule)
    return out


def link3(inp, descriptive=False, verbose=False):
    rule = rule_id2text["15"]
    out = inp

    pairs = [("ᆨ ᄋ", " ᄀ"),
             ("ᆩ ᄋ", " ᄁ"),
             ("ᆫ ᄋ", " ᄂ"),
             ("ᆮ ᄋ", " ᄃ"),
             ("ᆯ ᄋ", " ᄅ"),
             ("ᆷ ᄋ", " ᄆ"),
             ("ᆸ ᄋ", " ᄇ"),
             ("ᆺ ᄋ", " ᄉ"),
             ("ᆻ ᄋ", " ᄊ"),
             ("ᆽ ᄋ", " ᄌ"),
             ("ᆾ ᄋ", " ᄎ"),
             ("ᆿ ᄋ", " ᄏ"),
             ("ᇀ ᄋ", " ᄐ"),
             ("ᇁ ᄋ", " ᄑ"),

             ("ᆪ ᄋ", "ᆨ ᄊ"),
             ("ᆬ ᄋ", "ᆫ ᄌ"),
             ("ᆰ ᄋ", "ᆯ ᄀ"),
             ("ᆱ ᄋ", "ᆯ ᄆ"),
             ("ᆲ ᄋ", "ᆯ ᄇ"),
             ("ᆳ ᄋ", "ᆯ ᄊ"),
             ("ᆴ ᄋ", "ᆯ ᄐ"),
             ("ᆵ ᄋ", "ᆯ ᄑ"),
             ("ᆹ ᄋ", "ᆸ ᄊ")]

    for str1, str2 in pairs:
        out = out.replace(str1, str2)

    gloss(verbose, out, inp, rule)
    return out


def link4(inp, descriptive=False, verbose=False):
    rule = rule_id2text["12.4"]

    out = inp

    pairs = [("ᇂᄋ", "ᄋ"),
             ("ᆭᄋ", "ᄂ"),
             ("ᆶᄋ", "ᄅ")]

    for str1, str2 in pairs:
        out = out.replace(str1, str2)

    gloss(verbose, out, inp, rule)
    return out



# -*- coding:utf-8 -*-
# This was copied from [hanja](https://github.com/suminb/hanja) and modified by Kss

import os

import yaml


def load_table(filename):
    """Loads the Hanja table."""
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import BaseLoader as Loader
    with open(filename) as fin:
        table = yaml.load(fin.read(), Loader=Loader)

    return table


basepath = os.path.abspath(os.path.dirname(__file__))
hanja_table = load_table(os.path.join(basepath, "assets", "table.yml"))


def separate(ch):
    """한글 자모 분리. 주어진 한글 한 글자의 초성, 중성 초성을 반환함."""
    uindex = ord(ch) - 0xAC00
    jongseong = uindex % 28
    # NOTE: Force integer-divisions
    joongseong = ((uindex - jongseong) // 28) % 21
    choseong = ((uindex - jongseong) // 28) // 21

    return choseong, joongseong, jongseong


def synthesize(choseong, joongseong, jongseong):
    return build(choseong, joongseong, jongseong)


def build(choseong, joongseong, jongseong):
    """초성, 중성, 종성을 조합하여 완성형 한 글자를 만듦. 'choseong',
    'joongseong', 'jongseong' are offsets. For example, 'ㄱ' is 0, 'ㄲ' is 1,
    'ㄴ' is 2, and so on and so fourth."""
    code = int(((((choseong) * 21) + joongseong) * 28) + jongseong + 0xAC00)
    return chr(code)


def dooeum(previous, current):
    """두음법칙을 적용하기 위한 함수."""
    p, c = separate(previous), separate(current)
    offset = 0

    current_head = build(c[0], c[1], 0)

    # 모음이나 ㄴ 받침 뒤에 이어지는 '렬, 률'은 '열, 율'로 발음한다.
    if previous.isalnum():
        if current in ("렬", "률") and is_hangul(previous) and p[2] in (0, 2):
            offset = 6
    # 한자음 '녀, 뇨, 뉴, 니', '랴, 려, 례, 료, 류, 리'가 단어 첫머리에 올 때
    # '여, 요, 유, 이', '야, 여, 예, 요, 유, 이'로 발음한다.
    elif current_head in ("녀", "뇨", "뉴", "니"):
        offset = 9
    elif current_head in ("랴", "려", "례", "료", "류", "리"):
        offset = 6
    # 한자음 '라, 래, 로, 뢰, 루, 르'가 단어 첫머리에 올 때 '나, 내, 노, 뇌,
    # 누, 느'로 발음한다.
    elif current_head in ("라", "래", "로", "뢰", "루", "르"):
        offset = -3

    return build(c[0] + offset, c[1], c[2])


def is_hangul(ch):
    if ch is None:
        return False
    else:
        return 0xAC00 <= ord(ch) <= 0xD7A3


def contains_hangul(text):
    if isinstance(text, str):
        # NOTE: Probably not an ideal solution in terms of performance
        tfs = map(lambda c: is_hangul(c), text)
        for tf in tfs:
            if tf:
                return True
    return False


def translate_syllable(previous, current):
    """Translates a single syllable."""

    if current in hanja_table:
        return dooeum(previous, hanja_table[current])

    return current


def _split_hanja(text):
    """주어진 문장을 한자로 된 구역과 그 이외의 문자로 된 구역으로 분리"""

    # TODO: Can we make this a bit prettier?
    if len(text) == 0:
        yield text
    else:
        ch = text[0]
        bucket = [ch]
        prev_state = is_hanja(ch)

        for ch in text[1:]:
            state = is_hanja(ch)

            if prev_state != state:
                yield "".join(bucket)
                bucket = [ch]
            else:
                bucket.append(ch)

            prev_state = state

        yield "".join(bucket)


def split_hanja(text):
    """Splits a given text into hanja and non-hanja parts."""
    return list(_split_hanja(text))


def get_format_string(mode, word):
    """
    :param mode: substitution | combination-text | combination-text-reversed | combination-html | combination-html-reversed
    """
    if mode == "combination-text" and is_hanja(word[0]):
        return u"{word}({translated})"
    elif mode == "combination-text-reversed" and is_hanja(word[0]):
        return u"{translated}({word})"
    elif mode in "combination-html" and is_hanja(word[0]):
        return u'<span class="hanja">{word}</span><span class="hangul">({translated})</span>'
    elif mode in "combination-html-reversed" and is_hanja(word[0]):
        return u'<span class="hangul">{translated}</span><span class="hanja">({word})</span>'
    else:
        return u"{translated}"


def translate(text, mode):
    """Translates entire text."""
    words = list(split_hanja(text))
    return "".join(
        map(
            lambda w, prev: translate_word(w, prev, get_format_string(mode, w)),
            words,
            [None] + words[:-1],
        )
    )


def translate_word(word, prev, format_string):
    """Translates a single word.

    :param word: Word to be translated
    :param prev: Preceeding word
    """
    prev_char = prev[-1] if prev else " "
    buf = []
    for c in word:
        new_char = translate_syllable(prev_char, c)
        buf.append(new_char)
        prev_char = new_char
    translated = "".join(buf)

    return format_string.format(word=word, translated=translated)


def is_hanja(ch):
    """Determines if a given character ``ch`` is a Chinese character."""
    return ch in hanja_table



# Copyright (C) 2021 Hyunwoong Ko <kevin.brain@kakaobrain.com> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
import difflib
import logging
from logging import getLogger

logging.basicConfig(format="[Kss]: %(message)s", level=logging.WARNING)
logger = getLogger()


def highlight_diffs(old, new):
    red = lambda text: f"\033[38;2;255;0;0m{text}\033[0m"
    green = lambda text: f"\033[38;2;0;255;0m{text}\033[0m"

    result = ""
    codes = difflib.SequenceMatcher(a=old, b=new).get_opcodes()
    for code in codes:
        if code[0] == "equal":
            result += old[code[1]:code[2]]
        elif code[0] == "delete":
            result += red(old[code[1]:code[2]])
        elif code[0] == "insert":
            result += green(new[code[3]:code[4]])
        elif code[0] == "replace":
            result += (red(old[code[1]:code[2]]) + green(new[code[3]:code[4]]))
    return result


import re

from kss._modules.g2p.english import convert_eng
from kss._modules.g2p.numerals import convert_num

unicode_initial = [chr(initial_code) for initial_code in range(4352, 4371)]
unicode_medial = [
    'ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ'
]
unicode_final = [chr(final_code) for final_code in range(0x11a8, 0x11c3)]
unicode_final.insert(0, None)
unicode_offset = 44032
unicode_initial_offset = 588
unicode_medial_offset = 28

unicode_compatible_consonants = [
    'ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ'
]
unicode_compatible_finals = [
    'ᆨ', 'ᆩ', 'ᆫ', 'ᆮ', '_', 'ᆯ', 'ᆷ', 'ᆸ', '_', 'ᆺ', 'ᆻ', 'ᆼ', 'ᆽ', '_', 'ᆾ', 'ᆿ', 'ᇀ', 'ᇁ', 'ᇂ'
]

double_consonant_final = {
    'ᆪ': ('ᆨ', 'ᆺ'),
    'ᆬ': ('ᆫ', 'ᆽ'),
    'ᆭ': ('ᆫ', 'ᇂ'),
    'ᆰ': ('ᆯ', 'ᆨ'),
    'ᆱ': ('ᆯ', 'ᆷ'),
    'ᆲ': ('ᆯ', 'ᆸ'),
    'ᆳ': ('ᆯ', 'ᆻ'),
    'ᆴ': ('ᆯ', 'ᇀ'),
    'ᆵ': ('ᆯ', 'ᇁ'),
    'ᆶ': ('ᆯ', 'ᇂ'),
    'ᆹ': ('ᆸ', 'ᆺ'),
    'ㅆ': ('ㅅ', 'ㅅ')
}

NULL_CONSONANT = 'ᄋ'


class Syllable(object):
    def __init__(self, char):
        self.char = char
        _is_hangul, _separated = self.separate_syllable(char)
        if _is_hangul:
            self.initial = unicode_initial[_separated[0]]
            self.medial = unicode_medial[_separated[1]]
            self.final = unicode_final[_separated[2]]
        else:
            self.initial = _separated[0]
            self.medial = None
            self.final = None

    def separate_syllable(self, char):
        if self.is_hangul(char):
            initial = (ord(char) - unicode_offset) // unicode_initial_offset
            medial = ((ord(char) - unicode_offset) - unicode_initial_offset * initial) // unicode_medial_offset
            final = (((ord(char) - unicode_offset) - unicode_initial_offset * initial) - unicode_medial_offset * medial)
        else:
            initial = ord(char)
            medial = None
            final = None

        return self.is_hangul(char), [initial, medial, final]

    def construct_syllable(self, initial, medial, final):
        if self.is_hangul(self.char):
            initial = ord(initial) - 4352
            medial = unicode_medial.index(medial)
            if final is None:
                final = 0
            else:
                final = unicode_final.index(final)
            constructed = chr(
                (((initial * unicode_initial_offset) + (medial * unicode_medial_offset)) + final) + unicode_offset)
        else:
            constructed = self.char

        self.char = constructed
        return constructed

    @staticmethod
    def is_hangul(char):
        return 0xAC00 <= ord(char) <= 0xD7A3

    @staticmethod
    def final_to_initial(char):
        idx = unicode_compatible_finals.index(char)
        return unicode_initial[idx]

    def __repr__(self):
        self.construct_syllable(self.initial, self.medial, self.final)
        return self.char

    def __str__(self):
        self.char = self.construct_syllable(self.initial, self.medial, self.final)
        return self.char


def pronounce(
    text,
    convert_english_to_hangul_phonemes=False,
    convert_numbers_to_hangul_phonemes=False,
):
    if convert_english_to_hangul_phonemes:
        text = convert_eng(text)
    if convert_numbers_to_hangul_phonemes:
        text = convert_num(text)

    _syllables = [Syllable(char) for char in text]
    for idx, syllable in enumerate(_syllables):
        try:
            next_syllable = _syllables[idx + 1]
        except IndexError:
            next_syllable = None

        try:
            final_is_before_C = syllable.final and next_syllable.initial not in (
                None, NULL_CONSONANT)
        except AttributeError:
            final_is_before_C = False

        try:
            final_is_before_V = syllable.final and next_syllable.initial in (
                None, NULL_CONSONANT)
        except AttributeError:
            final_is_before_V = False

        is_last_syllable = syllable.final and next_syllable is None

        # 1. 받침 ‘ㄲ, ㅋ’, ‘ㅅ, ㅆ, ㅈ, ㅊ, ㅌ’, ‘ㅍ’은 어말 또는 자음 앞에서 각각 대표음 [ㄱ, ㄷ, ㅂ]으로 발음한다.
        # 2. 겹받침 ‘ㄳ’, ‘ㄵ’, ‘ㄼ, ㄽ, ㄾ’, ‘ㅄ’은 어말 또는 자음 앞에서 각각 [ㄱ, ㄴ, ㄹ, ㅂ]으로 발음한다.
        # 3. 겹받침 ‘ㄺ, ㄻ, ㄿ’은 어말 또는 자음 앞에서 각각 [ㄱ, ㅁ, ㅂ]으로 발음한다.
        # <-> 단, 국어의 로마자 표기법 규정에 의해 된소리되기는 표기에 반영하지 않으므로 제외.
        if is_last_syllable or final_is_before_C:
            if syllable.final in ['ᆩ', 'ᆿ', 'ᆪ', 'ᆰ']:
                syllable.final = 'ᆨ'
            elif syllable.final in ['ᆺ', 'ᆻ', 'ᆽ', 'ᆾ', 'ᇀ']:
                syllable.final = 'ᆮ'
            elif syllable.final in ['ᇁ', 'ᆹ', 'ᆵ']:
                syllable.final = 'ᆸ'
            elif syllable.final in ['ᆬ']:
                syllable.final = 'ᆫ'
            elif syllable.final in ['ᆲ', 'ᆳ', 'ᆴ']:
                syllable.final = 'ᆯ'
            elif syllable.final in ['ᆱ']:
                syllable.final = 'ᆷ'

        # 4. 받침 ‘ㅎ’의 발음은 다음과 같다.
        if syllable.final in ['ᇂ', 'ᆭ', 'ᆶ']:
            without_ㅎ = {
                'ᆭ': 'ᆫ',
                'ᆶ': 'ᆯ',
                'ᇂ': None
            }

            if next_syllable:
                # ‘ㅎ(ㄶ, ㅀ)’ 뒤에 ‘ㄱ, ㄷ, ㅈ’이 결합되는 경우에는, 뒤 음절 첫소리와 합쳐서 [ㅋ, ㅌ, ㅊ]으로 발음한다.
                # ‘ㅎ(ㄶ, ㅀ)’ 뒤에 ‘ㅅ’이 결합되는 경우에는, ‘ㅅ’을 [ㅆ]으로 발음한다.
                if next_syllable.initial in ['ᄀ', 'ᄃ', 'ᄌ', 'ᄉ']:
                    change_to = {'ᄀ': 'ᄏ', 'ᄃ': 'ᄐ', 'ᄌ': 'ᄎ', 'ᄉ': 'ᄊ'}
                    syllable.final = without_ㅎ[syllable.final]
                    next_syllable.initial = change_to[next_syllable.initial]
                # 3. ‘ㅎ’ 뒤에 ‘ㄴ’이 결합되는 경우에는, [ㄴ]으로 발음한다.
                elif next_syllable.initial in ['ᄂ']:
                    # TODO: [붙임] ‘ㄶ, ㅀ’ 뒤에 ‘ㄴ’이 결합되는 경우에는, ‘ㅎ’을 발음하지 않는다.
                    if syllable.final in ['ᆭ', 'ᆶ']:
                        syllable.final = without_ㅎ[syllable.final]
                    else:
                        syllable.final = 'ᆫ'
                # 4. ‘ㅎ(ㄶ, ㅀ)’ 뒤에 모음으로 시작된 어미나 접미사가 결합되는 경우에는,
                # ‘ㅎ’을 발음하지 않는다.
                elif next_syllable.initial == NULL_CONSONANT:
                    if syllable.final in ['ᆭ', 'ᆶ']:
                        if syllable.final == 'ᆭ':
                            syllable.final = 'ᆫ'
                        elif syllable.final == 'ᆶ':
                            syllable.final = 'ᆯ'
                    else:
                        syllable.final = None
                elif next_syllable.initial == 'ᄅ':
                    if syllable.final == 'ᆶ':
                        syllable.final = 'ᆯ'
                else:
                    if syllable.final == 'ᇂ':
                        syllable.final = None
            else:
                if syllable.final == 'ᇂ':
                    syllable.final = None

        # 6. 겹받침이 모음으로 시작된 조사나 어미, 접미사와 결합되는 경우에는,
        # 뒤엣것만을 뒤 음절 첫소리로 옮겨 발음한다.(이 경우, ‘ㅅ’은 된소리로 발음함.)
        if syllable.final in double_consonant_final and next_syllable.initial == NULL_CONSONANT:
            double_consonant = double_consonant_final[syllable.final]
            syllable.final = double_consonant[0]
            next_syllable.initial = next_syllable.final_to_initial(
                double_consonant[1])

        # 5. 홑받침이나 쌍받침이 모음으로 시작된 조사나 어미, 접미사와 결합되는 경우에는,
        # 제 음가대로 뒤 음절 첫소리로 옮겨 발음한다.
        if next_syllable and final_is_before_V:
            # do nothing if final is ᆼ or null
            if next_syllable.initial == NULL_CONSONANT and syllable.final not in ["ᆼ", None]:
                next_syllable.initial = next_syllable.final_to_initial(
                    syllable.final)
                syllable.final = None

    return ''.join([str(c) for c in _syllables])




# -*- coding: utf-8 -*-
"""
    hangulize.models
    ~~~~~~~~~~~~~~~~

    :copyright: (c) 2010-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
"""
from __future__ import absolute_import

from collections import deque
import functools
import re
import sys

from kss._modules.hangulization.hangulize import hangul
from kss._modules.hangulization.hangulize.hangul import join


__all__ = ['SPACE', 'ZWSP', 'EDGE', 'SPECIAL', 'BLANK', 'DONE', 'ENCODING',
           'EMPTY_TUPLE', 'Phoneme', 'Choseong', 'Jungseong', 'Jongseong',
           'Impurity', 'Notation', 'Language', 'Rewrite']


# include Hangul constants.
for name in dir(hangul):
    if name == name.upper():
        locals()[name] = getattr(hangul, name)
        __all__.append(name)


SPACE = ' '
ZWSP = '/'  # zero-width space
EDGE = chr(3)
SPECIAL = chr(6)
BLANK = '(?:%s)' % '|'.join(map(re.escape, (SPACE, ZWSP, EDGE, SPECIAL)))
DONE = chr(0)
ENCODING = getattr(sys.stdout, 'encoding', 'utf-8')
EMPTY_TUPLE = ()


def cached_property(func, name=None):
    if name is None:
        name = func.__name__
    def get(self):
        try:
            return self.__dict__[name]
        except KeyError:
            val = func(self)
            self.__dict__[name] = val
            return val
    functools.update_wrapper(get, func)
    def del_(self):
        self.__dict__.pop(name, None)
    return property(get, None, del_)


class Phoneme(object):
    """This abstract class wraps a Hangul letter."""

    def __init__(self, letter):
        self.letter = letter

    def __repr__(self):
        name = type(self).__name__
        return "<%s '%s'>" % (name, self.letter.encode(ENCODING))


class Choseong(Phoneme):
    """A initial consonant in Hangul::

        >>> Choseong(G)
        <Choseong 'ㄱ'>

    """

    pass


class Jungseong(Phoneme):
    """A vowel in Hangul::

        >>> Jungseong(A)
        <Jungseong 'ㅏ'>

    """

    pass


class Jongseong(Phoneme):
    """A final consonant in Hangul::

        >>> Jongseong(G)
        <Jongseong 'ㄱ'>

    """

    pass


class Impurity(Phoneme):
    """An impurity letter will be kept."""

    pass


class Notation(object):
    """Describes loanword orthography.

    :param rules: the rewrite rules as an ordered key-value list
    """

    def __init__(self, rules):
        self.rules = rules

    def __add__(self, rules):
        if isinstance(rules, Notation):
            rules = rules.rules
        return Notation(self.rules + rules)

    def __radd__(self, lrules):
        if isinstance(lrules, Notation):
            lrules = lrules.rules
        return Notation(lrules + self.rules)

    def __iter__(self):
        if not getattr(self, '_rewrites', None):
            self._rewrites = [Rewrite(*item) for item in self.items()]
        return iter(self._rewrites)

    def items(self, left_edge=False, right_edge=False, lang=None):
        """Yields each notation rules as regex."""
        for one in self.rules:
            pattern = one[0]
            # accept *args
            if len(one) == 2:
                val = one[1]
                if isinstance(val, Phoneme):
                    val = val,
            # accept args(a tuple instance)
            else:
                val = one[1:]
            yield pattern, val

    @property
    def chars(self):
        """The humane characters from the notation keys."""
        chest = []
        for one in self.rules:
            pattern = Rewrite.VARIABLE_PATTERN.sub('', one[0])
            pattern = re.sub(r'[\{\}\@\[\]\^\$]', '', pattern)
            for c in pattern:
                chest.append(c)
        return set(chest)


class Language(object):
    """Wraps a foreign language. The language should have a :class:`Notation`
    instance::

        >>> class Extraterrestrial(Language):
        ...     notation = Notation([
        ...         (u'ㅹ', (Choseong(BB), Jungseong(U), Jongseong(NG))),
        ...         (u'㉠', (Choseong(G),)),
        ...         (u'ㅣ', (Jungseong(I),)),
        ...         (u'ㅋ', (Choseong(K), Jungseong(I), Jongseong(G)))
        ...     ])
        ...
        >>> ext = Extraterrestrial()
        >>> print ext.hangulize(u'ㅹ㉠ㅣㅋㅋㅋ')
        뿡기킥킥킥

    :param logger: a logger

    """

    __tmp__ = ''
    __special__ = '.,;?~"()[]{}'

    vowels = EMPTY_TUPLE
    notation = None

    def __new__(cls):
        if not getattr(cls, '_instances', None):
            cls._instances = {}
        if cls not in cls._instances:
            cls._instances[cls] = object.__new__(cls)
        return cls._instances[cls]

    def __init__(self):
        if not isinstance(self.notation, Notation):
            raise NotImplementedError("notation has to be defined")

    @cached_property
    def _steal_specials(self):
        def keep(match, rewrite):
            """keep special characters."""
            self._specials.append(match.group(0))
            return SPECIAL
        esc = '(%s)' % '|'.join(re.escape(x) for x in
                                self.__special__ + self.__tmp__)
        return Rewrite(esc, keep)

    @cached_property
    def _recover_specials(self):
        def escape(match, rewrite):
            """escape special characters."""
            return (Impurity(self._specials.pop(0)),)
        return Rewrite(SPECIAL, escape)

    @cached_property
    def _remove_tmp(self):
        tmp = '(%s)' % '|'.join(re.escape(x) for x in self.__tmp__)
        return Rewrite(tmp, None)

    @property
    def chars_pattern(self):
        """The regex pattern which is matched the valid characters."""
        return ''.join(re.escape(c) for c in self.notation.chars)

    def split(self, string):
        """Splits words from the string. Each words have only valid characters.
        """
        pattern = '[^%s]+' % self.chars_pattern
        return re.split(pattern, string)

    def transcribe(self, string, logger=None):
        """Returns :class:`Phoneme` instance list from the word."""
        string = re.sub(r'\s+', SPACE, string)
        string = re.sub(r'^|$', EDGE, string)
        self._specials = []
        phonemes = []

        # steal special characters
        string = self._steal_specials(string, phonemes)
        # apply the notation
        for rewrite in self.notation:
            string = rewrite(string, phonemes, lang=self, logger=logger)
        # remove temporary characters
        string = self._remove_tmp(string, phonemes)
        # recover special characters
        string = self._recover_specials(string, phonemes)

        # post processing
        string = re.sub('^' + BLANK, '', string)
        string = re.sub(BLANK + '$', '', string)
        phonemes = phonemes[1:-1]
        string = _hold_spaces(string, phonemes)
        string = _remove_zwsp(string, phonemes)
        string = _pass_unmatched(string, phonemes)

        # flatten
        flat_phonemes = []
        for phs in phonemes:
            if phs:
                flat_phonemes.extend(phs)
        return flat_phonemes

    def normalize(self, string):
        """Before transcribing, normalizes the string. You could specify the
        different normalization for the language with overriding this method.
        """
        return string

    def hangulize(self, string, logger=None):
        """Hangulizes the string::

            >>> from kss._modules.hangulization.hangulize.langs.ja import Japanese
            >>> ja = Japanese()
            >>> ja.hangulize(u'あかちゃん')
            아카찬

        """
        from kss._modules.hangulization.hangulize.processing import complete_syllables
        def stringify(syllable):
            if isinstance(syllable[0], Impurity):
                return syllable[0].letter
            else:
                return join(syllable)
        string = self.normalize(string)
        logger and logger.info(">> '%s'" % string)
        phonemes = self.transcribe(string, logger=logger)
        try:
            syllables = complete_syllables(phonemes)
            result = [stringify(syl) for syl in syllables]
            hangulized = ''.join(result)
        except TypeError:
            hangulized = u''
        logger and logger.info('=> %s' % hangulized)
        return hangulized

    @property
    def iso639_1(self):
        return self.__iso639__.get(1)

    @property
    def iso639_2(self):
        return self.__iso639__.get(2)

    @property
    def iso639_3(self):
        return self.__iso639__.get(3)

    @property
    def code(self):
        return re.sub('^hangulize\.langs\.', '', type(self).__module__)


class Rewrite(object):

    VOWELS_PATTERN = re.compile('@')
    VARIABLE_PATTERN = re.compile('<(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)>')
    LEFT_EDGE_PATTERN = re.compile(r'^(\^?)\^')
    RIGHT_EDGE_PATTERN = re.compile(r'\$(\$?)$')
    LOOKBEHIND_PATTERN = re.compile('^(?P<edge>(?:\^(?:\^)?)?){([^}]+?)}')
    LOOKAHEAD_PATTERN = re.compile('{([^}]+?)}(?P<edge>(?:\$(?:\$)?)?)$')
    def NEGATIVE(regex):
        pattern = regex.pattern.replace('{', '{~')
        return re.compile(pattern)
    NEGATIVE_LOOKBEHIND_PATTERN = NEGATIVE(LOOKBEHIND_PATTERN)
    NEGATIVE_LOOKAHEAD_PATTERN = NEGATIVE(LOOKAHEAD_PATTERN)
    del NEGATIVE

    def __init__(self, pattern, val):
        """Makes a replace function with the given pattern and value."""
        self.pattern, self.val = pattern, val
        self.__regexes__ = {}

    def __call__(self, string, phonemes=None, lang=None, logger=None):
        # allocate needed offsets
        if not phonemes and isinstance(phonemes, list):
            phonemes += [None] * len(string)

        regex = self.compile_pattern(lang)

        # replacement function
        if phonemes:
            deletions = []
        def repl(match):
            val = self.val(match, self) if callable(self.val) else self.val
            repls.append(val)
            start, end = match.span()

            if val:
                is_tuple = isinstance(val, tuple)
                if not is_tuple:
                    if lang:
                        # variable replacement
                        cls = type(self)
                        srcvars = cls.find_actual_variables(self.pattern)
                        dstvars = cls.find_actual_variables(self.val)
                        srcvars, dstvars = list(srcvars), list(dstvars)
                        if len(srcvars) == len(dstvars) == 1:
                            src = getattr(lang, srcvars[0].group('name'))
                            dst = getattr(lang, dstvars[0].group('name'))
                            if len(src) != len(dst):
                                msg = 'the destination variable should ' \
                                      'have the same length with the ' \
                                      'source variable'
                                raise ValueError(msg)
                            dictionary = dict(zip(src, dst))
                            let = dictionary[match.group(0)]
                            val = self.VARIABLE_PATTERN.sub(let, val)
                        # group reference
                        val = re.sub(r'\\(\d+)',
                                     lambda m: match.group(int(m.group(1))),
                                     val)
                    if phonemes:
                        for x in range(len(val) - len(match.group(0))):
                            phonemes.insert(start, None)
                    return val
                elif phonemes and is_tuple:
                    # toss phonemes, and check the matched string
                    phonemes[start] = val
                    return DONE * (end - start)
            else:
                # when val is None, the matched string should remove
                if phonemes:
                    deletions.append((start, end))
                return ''

        if logger:
            prev = string
        repls = []

        # replace the string
        string = regex.sub(repl, string)

        # remove kept deletions
        if phonemes and deletions:
            _phonemes = deque()
            cur = len(phonemes)
            for start, end in reversed(deletions):
                _phonemes.extendleft(reversed(phonemes[end:cur]))
                cur = start
            _phonemes.extendleft(reversed(phonemes[0:cur]))
            phonemes[:] = _phonemes

        if logger:
            # report changes
            if prev != string:
                val = repls.pop()
                args = (string, self.pattern)
                if not val:
                    msg = ".. '%s'\tremove %s" % args
                elif isinstance(val, tuple):
                    val = ''.join(x.letter for x in val)
                    msg = ".. '%s'\thangulize %s -> %s" % (args + (val,))
                else:
                    msg = ".. '%s'\trewrite %s -> %s" % (args + (val,))
                logger.info(msg)

        return string

    def compile_pattern(self, lang=None):
        if lang not in self.__regexes__:
            regex = re.compile(type(self).regexify(self.pattern, lang))
            self.__regexes__[lang] = regex
        return self.__regexes__[lang]

    @classmethod
    def regexify(cls, pattern, lang=None):
        regex = pattern
        if lang:
            regex = cls.regexify_variable(regex, lang)
        regex = cls.regexify_negative_lookaround(regex)
        regex = cls.regexify_lookaround(regex)
        regex = cls.regexify_edge_of_word(regex)
        return regex

    @classmethod
    def regexify_edge_of_word(cls, regex):
        left_edge = r'(?<=\1%s)' % BLANK
        right_edge = r'(?=%s\1)' % BLANK
        regex = cls.LEFT_EDGE_PATTERN.sub(left_edge, regex)
        regex = cls.RIGHT_EDGE_PATTERN.sub(right_edge, regex)
        return regex

    def _make_lookaround(behind_pattern, ahead_pattern,
                         behind_prefix, ahead_prefix):
        @staticmethod
        def meth(regex):
            def lookbehind(match):
                edge = re.sub('\^$', BLANK, match.group('edge'))
                return '(?' + behind_prefix + edge + \
                       '(?:' + match.group(2) + '))'
            def lookahead(match):
                edge = re.sub('^\$', BLANK, match.group('edge'))
                return '(?' + ahead_prefix + \
                       '(?:' + match.group(1) + ')' + edge + ')'
            regex = behind_pattern.sub(lookbehind, regex)
            regex = ahead_pattern.sub(lookahead, regex)
            return regex
        return meth

    _positive = _make_lookaround(LOOKBEHIND_PATTERN,
                                 LOOKAHEAD_PATTERN,
                                 '<=', '=')
    _negative = _make_lookaround(NEGATIVE_LOOKBEHIND_PATTERN,
                                 NEGATIVE_LOOKAHEAD_PATTERN,
                                 '<!', '!')
    regexify_lookaround = _positive
    regexify_negative_lookaround = _negative
    del _make_lookaround, _positive, _negative

    @classmethod
    def regexify_variable(cls, regex, lang):
        def to_variable(match):
            var = getattr(lang, match.group('name'))
            return '(%s)' % '|'.join(re.escape(x) for x in var)
        regex = cls.VOWELS_PATTERN.sub('<vowels>', regex)
        regex = cls.VARIABLE_PATTERN.sub(to_variable, regex)
        return regex

    @classmethod
    def find_actual_variables(cls, pattern):
        # pass when there's no any variable patterns
        if not cls.VOWELS_PATTERN.search(pattern) and \
           not cls.VARIABLE_PATTERN.search(pattern):
            return EMPTY_TUPLE
        try:
            pattern = cls.LOOKBEHIND_PATTERN.sub(DONE, pattern)
            pattern = cls.LOOKAHEAD_PATTERN.sub(DONE, pattern)
            pattern = cls.VOWELS_PATTERN.sub('<>', pattern)
            return cls.VARIABLE_PATTERN.finditer(pattern)
        except TypeError:
            return EMPTY_TUPLE


_remove_zwsp = Rewrite(ZWSP, (Impurity(''),))
_hold_spaces = Rewrite(SPACE, (Impurity(' '),))
_pass_unmatched = Rewrite('[^' + DONE + ']+',
                          lambda m, r: (Impurity(m.group(0)),))



# This was copied from KoEDA [https://github.com/toriving/KoEDA]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import random
from itertools import repeat, chain
from typing import Union, List

from kss._modules.morphemes.split_morphemes import split_morphemes
from kss._modules.augmentation.utils import get_synonyms
from kss._modules.augmentation.distance import hangul_levenshtein


class SynonymReplacement:
    def __init__(self, backend):
        self.backend = backend

    def __call__(self, *args, **kwargs):
        return self.synonym_replacement(*args, **kwargs)

    def synonym_replacement(
        self, data: Union[List[str], str], p: float = 0.1, repetition: int = 1, verbose: bool = False
    ) -> Union[List[str], str]:
        if isinstance(data, str):
            if repetition <= 1:
                return self._replacement(data, p, verbose)
            else:
                return list(
                    map(
                        self._replacement,
                        repeat(data, repetition),
                        repeat(p, repetition),
                        repeat(verbose, repetition),
                    )
                )
        elif isinstance(data, list):
            if repetition <= 1:
                return list(map(self._replacement, data, repeat(p, len(data)), repeat(verbose, len(data))))
            else:
                return list(
                    map(
                        self._replacement,
                        chain.from_iterable(repeat(x, repetition) for x in data),
                        repeat(p, len(data) * repetition),
                        repeat(verbose, len(data) * repetition),
                    )
                )
        else:
            raise TypeError(f"Does not support the data type : {type(data)}")

    def _replacement(self, data: str, p: float = 0.1, verbose: bool=False) -> str:
        words = split_morphemes(data, backend=self.backend, drop_space=False)
        n = max(0, int(len(words) * p))

        random_word_list = list(set([
            word for word, pos in words
            if pos.startswith("N") and len(word) > 1
        ]))

        random.shuffle(random_word_list)
        num_replaced = 0
        old2new = {}
        for random_word in random_word_list:
            synonyms = get_synonyms(random_word, min_length=2)

            if len(synonyms) >= 1:
                synonym2distance = {}
                for synonym in synonyms:
                    distance = hangul_levenshtein(synonym, random_word)
                    synonym2distance[synonym] = distance

                closet_synonym = min(synonym2distance, key=synonym2distance.get)
                closet_synonym_distance = synonym2distance[closet_synonym]
                if closet_synonym_distance > 0.5:
                    continue
                for word, pos in words:
                    if word == random_word:
                        old2new[word] = closet_synonym
                        break
                num_replaced += 1
            if num_replaced >= n:
                break

        finalized_words = []
        verbose_dict = {}

        for i, (word, pos) in enumerate(words):
            if word in old2new and pos.startswith("N"):
                finalized_words.append(old2new[word])
                verbose_dict[word] = old2new[word]
            else:
                finalized_words.append(word)

        if verbose:
            for word, replaced_word in verbose_dict.items():
                print(f"{word} -> {replaced_word}")

        return "".join(finalized_words)
